import re
import unicodedata
from pathlib import Path
from typing import NamedTuple
from zipfile import ZipInfo

from bs4 import BeautifulSoup, Tag

from labels.model.ecosystem_data.java import (
    JavaArchive,
    JavaManifest,
    JavaPomParent,
    JavaPomProject,
    JavaPomProperties,
)
from labels.model.file import Location
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.parsers.cataloger.java.parse_archive.archive_file_info import ArchiveFileInfo
from labels.parsers.cataloger.java.utils.package_builder import new_java_package_from_maven_data
from labels.parsers.cataloger.java.utils.parsing_utils import get_direct_child_text, safe_get_text
from labels.parsers.cataloger.utils import get_enriched_location
from labels.utils.zip import contents_from_zip, new_zip_glob_match


class JavaPackageIdentity(NamedTuple):
    artifact_id: str | None
    version: str | None
    group_id: str | None


class MavenHandler:
    def __init__(
        self,
        *,
        file_manifest: list[ZipInfo],
        archive_path: str,
        location: Location,
        file_info: ArchiveFileInfo,
    ) -> None:
        self.file_manifest = file_manifest
        self.archive_path = archive_path
        self.location = location
        self.file_info = file_info
        self._properties_cache: dict[str, JavaPomProperties] | None = None
        self._projects_cache: dict[str, JavaPomProject] | None = None

    def discover_auxiliary_packages(self, parent_package: Package) -> list[Package]:
        properties = self._load_properties()
        projects = self._load_projects()

        packages = []
        for parent_path, properties_obj in properties.items():
            pom_project = projects.get(parent_path)

            new_location = get_enriched_location(
                self.location, package_manager=PackageManager.MAVEN
            )

            package = new_java_package_from_maven_data(properties_obj, pom_project, new_location)
            if package and not self._same_as_parent_package(parent_package, package):
                packages.append(package)

        return packages

    def extract_identity(self, manifest: JavaManifest) -> JavaPackageIdentity:
        properties = self._load_properties()
        projects = self._load_projects()
        properties_obj, project_obj = self._find_matching_objects(properties, projects)

        artifact_id = None
        version = None
        group_id = None

        if properties_obj:
            artifact_id = properties_obj.artifact_id
            version = properties_obj.version
            group_id = properties_obj.group_id

        if project_obj and not version:
            version = project_obj.version
        if project_obj and not group_id:
            group_id = project_obj.group_id

        artifact_id = artifact_id or self._select_name_from_manifest_or_file(manifest)
        version = version or self._select_version_from_manifest_or_file(manifest)
        group_id = group_id or self._select_name_from_manifest_or_file(manifest)

        return JavaPackageIdentity(artifact_id=artifact_id, version=version, group_id=group_id)

    def build_ecosystem_data(self, manifest: JavaManifest) -> JavaArchive:
        properties = self._load_properties()
        projects = self._load_projects()
        props_obj, proj_obj = self._find_matching_objects(properties, projects)

        return JavaArchive(
            manifest=manifest,
            pom_properties=props_obj,
            pom_project=proj_obj if proj_obj else None,
        )

    def _same_as_parent_package(self, parent_package: Package, package: Package) -> bool:
        return package.name == parent_package.name and package.version == parent_package.version

    def _find_matching_objects(
        self, properties: dict[str, JavaPomProperties], projects: dict[str, JavaPomProject]
    ) -> tuple[JavaPomProperties | None, JavaPomProject | None]:
        artifacts_map = {props.artifact_id for props in properties.values() if props.artifact_id}

        for parent_path, properties_obj in properties.items():
            if properties_obj.artifact_id:
                artifact_matches_filename = self._artifact_id_matches_filename(
                    properties_obj.artifact_id, self.file_info.name, artifacts_map
                )
                project_object = projects.get(parent_path)
                if artifact_matches_filename and project_object:
                    return properties_obj, project_object

        return None, None

    def _artifact_id_matches_filename(
        self, artifact_id: str, filename: str, artifacts_map: set[str]
    ) -> bool:
        if filename in artifacts_map:
            return artifact_id == filename

        return artifact_id.startswith(filename) or filename.endswith(artifact_id)

    def _select_name_from_manifest_or_file(self, manifest: JavaManifest) -> str:
        name = self._extract_name_from_apache_maven_bundle_plugin(manifest)
        if name:
            return name

        # the filename tends to be the next-best reference for the package name
        name = self._extract_name_from_file_info()
        if name:
            return name

        # remaining fields in the manifest is a bit of a free-for-all depending on
        # the build tooling used and package maintainer preferences
        main_attrs = manifest.main
        field_names = [
            "Name",
            "Bundle-Name",
            "Short-Name",
            "Extension-Name",
            "Implementation-Title",
        ]
        for key in field_names:
            if main_attrs.get(key):
                return main_attrs[key]

        return ""

    def _extract_name_from_apache_maven_bundle_plugin(self, manifest: JavaManifest) -> str:
        created_by = manifest.main.get("Created-By", "")
        if "Apache Maven Bundle Plugin" in created_by:
            symbolic_name = manifest.main.get("Bundle-SymbolicName", "")
            if symbolic_name:
                vendor_id = manifest.main.get("Implementation-Vendor-Id", "")
                if vendor_id and vendor_id == symbolic_name:
                    return ""

                # Assuming symbolicName convention "${groupId}.${artifactId}".
                fields = symbolic_name.split(".")
                # Potential issue with determining the actual artifactId based
                # on BND behavior.
                return fields[-1] if fields else ""

        return ""

    def _extract_name_from_file_info(self) -> str:
        if "." in self.file_info.name:
            # Handle the special case for 'org.eclipse.*' group IDs
            if self.file_info.name.startswith("org.eclipse."):
                return self.file_info.name

            # Split the name on dots and check if it looks like a
            # 'groupId.artifactId'
            fields = self.file_info.name.split(".")
            if all(self._is_valid_java_identifier(f) for f in fields):
                # If all parts are valid Java identifiers, assume the last part
                # is the artifact ID
                return fields[-1]

        return self.file_info.name

    @staticmethod
    def _is_valid_java_identifier(field: str) -> bool:
        if not field:
            return False

        first_char = field[0]
        if first_char.isalpha() or unicodedata.category(first_char) in [
            "Sc",
            "Pc",
        ]:
            # If the first character is valid, check the remaining characters.
            # Note: Python's str.isidentifier() checks if the entire string is a
            # valid identifier.
            return field.isidentifier()

        return False

    def _select_version_from_manifest_or_file(self, manifest: JavaManifest) -> str:
        if version := self.file_info.version:
            return version

        field_names = [
            "Implementation-Version",
            "Specification-Version",
            "Plugin-Version",
            "Bundle-Version",
        ]
        for field in field_names:
            if value := self._field_value_from_manifest(manifest, field):
                return value

        return ""

    def _field_value_from_manifest(self, manifest: JavaManifest, field: str) -> str:
        if (value := manifest.main.get(field, None)) and value:
            return value

        for section in manifest.sections or []:
            if (value := section.get(field, None)) and value:
                return value
        return ""

    def _load_properties(self) -> dict[str, JavaPomProperties]:
        if self._properties_cache is None:
            matches = new_zip_glob_match(
                self.file_manifest, ("*pom.properties",), case_sensitive=False
            )
            self._properties_cache = self._pom_properties_by_parent(self.archive_path, matches)
        return self._properties_cache or {}

    def _pom_properties_by_parent(
        self, archive_path: str, extract_paths: list[str]
    ) -> dict[str, JavaPomProperties]:
        properties_by_parent_path = {}
        contents_of_maven_properties = contents_from_zip(archive_path, *extract_paths)

        for file_path, file_contents in contents_of_maven_properties.items():
            if not file_contents:
                continue

            pom_properties = self._parse_pom_properties(file_contents)

            properties_by_parent_path[str(Path(file_path).parent)] = pom_properties

        return properties_by_parent_path

    def _parse_pom_properties(self, file_content: str) -> JavaPomProperties:
        properties_map = {}

        for raw_line in file_content.splitlines():
            line = raw_line.strip()

            if line == "" or line.lstrip().startswith("#"):
                continue

            idx = next((i for i in range(len(line)) if line[i] in ":="), -1)
            if idx == -1:
                continue

            key = line[:idx].strip()
            value = line[idx + 1 :].strip()
            properties_map[key] = value

        converted_props = {}
        for raw_key, value in properties_map.items():
            key = re.sub(r"(?<!^)(?=[A-Z])", "_", raw_key).lower()
            if key in set(JavaPomProperties.__annotations__.keys()):
                converted_props[key] = value

        return JavaPomProperties(**converted_props)

    def _load_projects(self) -> dict[str, JavaPomProject]:
        if self._projects_cache is None:
            matches = new_zip_glob_match(self.file_manifest, ("*pom.xml",), case_sensitive=False)
            self._projects_cache = self._pom_project_by_parent(self.archive_path, matches)
        return self._projects_cache or {}

    def _pom_project_by_parent(
        self, archive_path: str, extract_paths: list[str]
    ) -> dict[str, JavaPomProject]:
        contents_of_maven_project = contents_from_zip(archive_path, *extract_paths)

        project_by_parent = {}

        for file_path, file_contents in contents_of_maven_project.items():
            pom_project = self._parse_pom_xml_project(file_contents)
            if not pom_project:
                continue

            project_by_parent[str(Path(file_path).parent)] = pom_project

        return project_by_parent

    def _parse_pom_xml_project(self, reader: str) -> JavaPomProject | None:
        project = BeautifulSoup(reader, features="xml").project
        if not project:
            return None

        return self._build_pom_project(project)

    def _build_pom_project(self, project: Tag) -> JavaPomProject:
        artifact_id = safe_get_text(self._find_direct_child(project, "artifactId"))
        name = safe_get_text(self._find_direct_child(project, "name"))

        return JavaPomProject(
            parent=self._build_pom_parent(self._find_direct_child(project, "parent")),
            group_id=safe_get_text(self._find_direct_child(project, "groupId")),
            artifact_id=artifact_id,
            version=safe_get_text(self._find_direct_child(project, "version")),
            name=name,
        )

    def _build_pom_parent(self, parent: Tag | None) -> JavaPomParent | None:
        if not parent:
            return None

        group_id = get_direct_child_text(parent, "groupId")
        artifact_id = get_direct_child_text(parent, "artifactId")
        version = get_direct_child_text(parent, "version")

        if group_id is None or artifact_id is None or version is None:
            return None
        if not (group_id.strip() and artifact_id.strip() and version.strip()):
            return None

        return JavaPomParent(
            group_id=group_id,
            artifact_id=artifact_id,
            version=version,
        )

    def _find_direct_child(self, parent: Tag, tag: str) -> Tag | None:
        children: list[Tag] = parent.find_all(tag, recursive=False)
        return next(
            (child for child in children if child.parent == parent),
            None,
        )



import json
import sqlite3

from fluidattacks_core.semver.match_versions import match_vulnerable_versions

from labels.advisories.database import BaseDatabase
from labels.advisories.match_fixes import (
    match_fixes,
)
from labels.advisories.utils import create_advisory_from_record
from labels.model.advisories import Advisory, AdvisoryRecord


class RootsDatabase(BaseDatabase):
    def __init__(self) -> None:
        super().__init__(db_name="skims_sca_advisories.db")


DATABASE = RootsDatabase()


def fetch_advisory_from_database(
    cursor: sqlite3.Cursor,
    ecosystem: str,
    package_name: str,
) -> list[AdvisoryRecord]:
    cursor.execute(
        """
        SELECT
            adv_id,
            source,
            vulnerable_version,
            severity_level,
            severity_v4,
            epss,
            details,
            percentile,
            cwe_ids,
            cve_finding,
            auto_approve,
            fixed_versions,
            kev_catalog
        FROM advisories
        WHERE ecosystem = ? AND package_name = ?;
        """,
        (ecosystem, package_name),
    )
    records: list[AdvisoryRecord] = cursor.fetchall()
    return records


def fetch_safe_versions_from_database(
    cursor: sqlite3.Cursor,
    ecosystem: str,
    package_name: str,
) -> list[str]:
    cursor.execute(
        """
        SELECT safe_versions
        FROM safe_versions
        WHERE ecosystem = ? AND package_name = ?;
        """,
        (ecosystem, package_name),
    )
    row: tuple[str] | None = cursor.fetchone()
    if row and row[0]:
        result: list[str] = json.loads(row[0])
        return result
    return []


def get_package_advisories(
    ecosystem: str,
    package_name: str,
    version: str,
) -> list[Advisory]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()

    return [
        adv
        for advisory_record in fetch_advisory_from_database(cursor, ecosystem, package_name)
        if (adv := create_advisory_from_record(advisory_record, ecosystem, package_name, version))
    ]


def fetch_all_package_versions_from_database(
    cursor: sqlite3.Cursor,
    ecosystem: str,
    package_name: str,
) -> list[str]:
    cursor.execute(
        """
        SELECT available_versions
        FROM package_versions_registry
        WHERE ecosystem = ? AND package_name = ?;
        """,
        (ecosystem, package_name),
    )
    row: tuple[str] | None = cursor.fetchone()
    if row and row[0]:
        result: list[str] = json.loads(row[0])
        return result
    return []


def get_all_package_versions(
    ecosystem: str,
    package_name: str,
) -> list[str]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()
    return fetch_all_package_versions_from_database(cursor, ecosystem, package_name)


def get_safe_versions(
    ecosystem: str,
    package_name: str,
) -> list[str]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()
    return fetch_safe_versions_from_database(cursor, ecosystem, package_name)


def get_vulnerabilities(
    ecosystem: str, product: str, version: str, safe_versions: list[str] | None
) -> list[Advisory]:
    if (
        product
        and version
        and (advisories := get_package_advisories(ecosystem, product.lower(), version))
    ):
        all_package_versions = get_all_package_versions(ecosystem, product.lower())
        return [
            match_fixes(version.lower(), advisor, safe_versions, all_package_versions, advisories)
            for advisor in advisories
            if match_vulnerable_versions(version.lower(), advisor.vulnerable_version)
        ]
    return []



import logging

import tree_sitter_ruby
from tree_sitter import Language as TLanguage
from tree_sitter import Node, Parser, Query, QueryCursor

LOGGER = logging.getLogger(__name__)


def parse_ruby_with_tree_sitter(content: str) -> Node | None:
    try:
        parser_language = TLanguage(tree_sitter_ruby.language())
        parser = Parser(parser_language)
        result = parser.parse(content.encode("utf-8"))
    except Exception:
        LOGGER.exception("Failed to parse Ruby content with tree-sitter")
        return None
    else:
        return result.root_node if result.root_node else None


def find_method_calls(node: Node, method_name: str) -> list[Node]:
    try:
        language = TLanguage(tree_sitter_ruby.language())

        query_string = f"""
        (call
          method: (identifier) @method
          (#eq? @method "{method_name}")
        ) @call
        """

        query = Query(language, query_string)
        cursor = QueryCursor(query)

        captures = cursor.captures(node)

        return captures.get("call", [])
    except Exception:
        LOGGER.exception(
            "Failed to find method calls for method '%s'",
            method_name,
            extra={"method_name": method_name},
        )
        return []



import re
from typing import NamedTuple

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProject
from labels.model.file import (
    Location,
    LocationReadCloser,
)
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import (
    Relationship,
)
from labels.model.release import Environment
from labels.model.resolver import (
    Resolver,
)
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location

QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r'[^"\']+'
SBT_DEPENDENCY_RE = re.compile(
    r"^[^%]*"
    rf"{NL}{QUOTE}(?P<group>{TEXT}){QUOTE}{NL}%"
    rf"{NL}{QUOTE}(?P<name>{TEXT}){QUOTE}{NL}%"
    rf"{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}"
    r".*$",
)

VERSION_VAR_RE = re.compile(
    rf"^\s*val\s+(?P<var_name>\w+)(?:\s*:\s*\w+)?\s*=\s*{QUOTE}(?P<value>{TEXT}){QUOTE}\s*$",
)

VERSION_REF_RE = re.compile(
    rf"{QUOTE}(?P<group>{TEXT}){QUOTE}\s*%%?\s*{QUOTE}(?P<name>{TEXT}){QUOTE}\s*%\s*(?P<version_var>\w+)",
)


class SbtDependency(NamedTuple):
    group: str
    name: str
    version: str
    line_no: int


def parse_build_sbt(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    sbt_dependencies = _parse_dependencies(reader.read_closer.read())

    packages = _collect_packages(sbt_dependencies, reader.location)

    return packages, []


def _parse_dependencies(content: str) -> list[SbtDependency]:
    dependencies: list[SbtDependency] = []

    content_lines = content.splitlines()
    version_vars = _extract_version_vars(content_lines)

    for line_no, line in enumerate(content_lines, start=1):
        if match := SBT_DEPENDENCY_RE.match(line):
            version = match.group("version")
            dependencies.append(
                SbtDependency(
                    group=match.group("group"),
                    name=match.group("name"),
                    version=version,
                    line_no=line_no,
                )
            )
        elif match := VERSION_REF_RE.search(line):
            version_var = match.group("version_var")
            version = version_vars.get(version_var, "unknown")
            dependencies.append(
                SbtDependency(
                    group=match.group("group"),
                    name=match.group("name"),
                    version=version,
                    line_no=line_no,
                )
            )

    return dependencies


def _extract_version_vars(content: list[str]) -> dict[str, str]:
    version_vars = {}

    for line in content:
        if match := VERSION_VAR_RE.match(line):
            version_vars[match.group("var_name")] = match.group("value")

    return version_vars


def _collect_packages(sbt_dependencies: list[SbtDependency], location: Location) -> list[Package]:
    packages: list[Package] = []

    for dependency in sbt_dependencies:
        product = f"{dependency.group}:{dependency.name}"

        new_location = get_enriched_location(
            location,
            line=dependency.line_no,
            is_transitive=False,
            package_manager=PackageManager.SBT,
        )

        ecosystem_data = JavaArchive(
            pom_project=JavaPomProject(
                name=product,
                group_id=dependency.group,
                artifact_id=dependency.name,
                version=dependency.version,
            )
        )

        package_spec = JavaPackageSpec(
            simple_name=dependency.name,
            version=dependency.version,
            location=new_location,
            ecosystem_data=ecosystem_data,
            composed_name=product,
        )

        package = new_java_package(package_spec)
        if package:  # pragma: no branch
            packages.append(package)

    return packages



from collections import UserDict, UserList
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict
from tree_sitter import Node

KEY = TypeVar("KEY")
VALUE = TypeVar("VALUE")


class FileCoordinate(BaseModel):
    line: int
    column: int
    model_config = ConfigDict(frozen=True)


class Position(BaseModel):
    start: FileCoordinate
    end: FileCoordinate
    model_config = ConfigDict(frozen=True)


class IndexedDict(UserDict[KEY, VALUE], Generic[KEY, VALUE]):
    def __init__(self, root_node: Node | None = None) -> None:
        self.position_value_index: dict[KEY, Position] = {}
        self.position_key_index: dict[KEY, Position] = {}
        data: dict[KEY, VALUE] = {}
        if root_node:
            self.position = Position(
                start=FileCoordinate(
                    line=root_node.start_point[0] + 1,
                    column=root_node.start_point[1] + 1,
                ),
                end=FileCoordinate(
                    line=root_node.end_point[0] + 1,
                    column=root_node.end_point[1] + 1,
                ),
            )
        super().__init__(data)

    def __setitem__(  # type: ignore[override]
        self,
        key: tuple[KEY, Position | Node],
        item: tuple[VALUE, Position | Node],
    ) -> None:
        key_value, key_position = key
        value_value, value_position = item
        if isinstance(key_position, Node):
            key_position = Position(
                start=FileCoordinate(
                    line=key_position.start_point[0] + 1,
                    column=key_position.start_point[1] + 1,
                ),
                end=FileCoordinate(
                    line=key_position.end_point[0] + 1,
                    column=key_position.end_point[1] + 1,
                ),
            )
        if isinstance(value_position, Node):
            value_position = Position(
                start=FileCoordinate(
                    line=value_position.start_point[0] + 1,
                    column=value_position.start_point[1] + 1,
                ),
                end=FileCoordinate(
                    line=value_position.end_point[0] + 1,
                    column=value_position.end_point[1] + 1,
                ),
            )
        self.position_key_index[key_value] = key_position
        self.position_value_index[key_value] = value_position
        return super().__setitem__(key_value, value_value)

    def get_value_position(self, key: KEY) -> Position:
        return self.position_value_index[key]

    def get_key_position(self, key: KEY) -> Position:
        return self.position_key_index[key]


class IndexedList(UserList[VALUE], Generic[VALUE]):
    def __init__(self, node: Node | None = None) -> None:
        self.position_index: dict[int, Position] = {}
        data: list[VALUE] = []
        if node:
            self.position = Position(
                start=FileCoordinate(
                    line=node.start_point[0] + 1,
                    column=node.start_point[1] + 1,
                ),
                end=FileCoordinate(
                    line=node.end_point[0] + 1,
                    column=node.end_point[1] + 1,
                ),
            )
        super().__init__(data)

    def __setitem__(self, index: int, value: tuple[VALUE, Position]) -> None:  # type: ignore[override]
        self.position_index[index] = value[1]
        return super().__setitem__(index, value[0])

    def append(self, item: tuple[VALUE, Position | Node]) -> None:  # type: ignore[override]
        value, position = item
        if isinstance(position, Node):
            position = Position(
                start=FileCoordinate(
                    line=position.start_point[0] + 1,
                    column=position.start_point[1] + 1,
                ),
                end=FileCoordinate(
                    line=position.end_point[0] + 1,
                    column=position.end_point[1] + 1,
                ),
            )
        self.position_index[len(self.data)] = position
        return super().append(value)

    def get_position(self, index: int) -> Position:
        return self.position_index[index]


ParsedValue = (
    str
    | int
    | float
    | bool
    | None
    | Position
    | list[str]
    | IndexedDict[str, "ParsedValue"]
    | IndexedList["ParsedValue"]
)



from typing import TypeAlias

from labels.model.ecosystem_data.alpine import ApkDBEntry
from labels.model.ecosystem_data.arch import AlpmDBEntry
from labels.model.ecosystem_data.debian import DpkgDBEntry
from labels.model.ecosystem_data.java import JavaArchive
from labels.model.ecosystem_data.python import WheelEggEcosystemData
from labels.model.ecosystem_data.redhat import RpmDBEntry

AcceptedEcosystemData: TypeAlias = (
    AlpmDBEntry | ApkDBEntry | DpkgDBEntry | JavaArchive | RpmDBEntry | WheelEggEcosystemData
)



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.dotnet.parse_ascx_scripts import parse_ascx_scripts
from labels.parsers.cataloger.dotnet.parse_csproj import parse_csproj
from labels.parsers.cataloger.dotnet.parse_dotnet_deps_json import parse_dotnet_deps_json
from labels.parsers.cataloger.dotnet.parse_dotnet_exe_config import parse_dotnet_config_executable
from labels.parsers.cataloger.dotnet.parse_dotnet_package_config import parse_dotnet_pkgs_config
from labels.parsers.cataloger.dotnet.parse_dotnet_package_lock import parse_dotnet_package_lock
from labels.parsers.cataloger.dotnet.parse_dotnet_portable_executable import (
    parse_dotnet_portable_executable,
)


def on_next_dotnet(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            patterns = [
                (
                    ("**/packages.config", "packages.config"),
                    parse_dotnet_pkgs_config,
                    "dotnet-parse-packages-config",
                ),
                (
                    ("**/packages.lock.json", "packages.lock.json"),
                    parse_dotnet_package_lock,
                    "dotnet-parse-package-lock",
                ),
                (
                    ("**/*.csproj", "*.csproj"),
                    parse_csproj,
                    "dotnet-parse-csproj",
                ),
                (
                    ("**/*.deps.json", "*.deps.json"),
                    parse_dotnet_deps_json,
                    "dotnet-parse-deps-json",
                ),
                (
                    ("**/*.dll", "*.dll", "**/*.exe", "*.exe"),
                    parse_dotnet_portable_executable,
                    "dotnet-parse-portable-executable",
                ),
                (
                    ("**/*.exe.config", "*.exe.config"),
                    parse_dotnet_config_executable,
                    "dotnet-parse-config-executable",
                ),
                (
                    ("**/*.ascx", "*.ascx"),
                    parse_ascx_scripts,
                    "dotnet-parse-ascx-scripts",
                ),
            ]
            try:
                for pattern, parser, parser_name in patterns:
                    if any(fnmatch(value, x) for x in pattern):
                        observer.on_next(
                            Request(
                                real_path=value,
                                parser=parser,
                                parser_name=parser_name,
                            ),
                        )
                        break
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from typing import NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.swift.package_builder import new_cocoa_pods_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


class PodInfo(NamedTuple):
    name: str
    version: str
    dependencies: list[str]
    is_transitive: bool
    line: int


def parse_podfile_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = _safe_parse_podfile(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    pods_info = _parse_pods_info(file_content)
    if not pods_info:
        return [], []

    packages = _collect_packages(pods_info, reader.location)
    relationships = _generate_relations(pods_info, packages)

    return packages, relationships


def _safe_parse_podfile(content: str) -> ParsedValue | None:
    try:
        parsed_podfile = parse_yaml_with_tree_sitter(content)
    except ValueError:
        return None

    return parsed_podfile


def _parse_pods_info(file_content: IndexedDict[str, ParsedValue]) -> list[PodInfo]:
    pods_info: list[PodInfo] = []

    direct_dependencies = file_content.get("DEPENDENCIES")
    pods = file_content.get("PODS")
    if not isinstance(direct_dependencies, IndexedList) or not isinstance(pods, IndexedList):
        return pods_info

    for index, pod in enumerate(pods):
        pod_name, pod_version = _extract_pod_info(pod)
        pod_dependencies = _extract_pod_dependencies(pod)
        is_transitive = pod_name not in direct_dependencies
        line = pods.get_position(index).start.line

        pods_info.append(
            PodInfo(
                name=pod_name,
                version=pod_version,
                dependencies=pod_dependencies,
                is_transitive=is_transitive,
                line=line,
            )
        )

    return pods_info


def _collect_packages(pods_info: list[PodInfo], location: Location) -> list[Package]:
    packages: list[Package] = []

    for pod_info in pods_info:
        new_location = get_enriched_location(
            location,
            is_transitive=pod_info.is_transitive,
            line=pod_info.line,
            package_manager=PackageManager.COCOAPODS,
        )

        package = new_cocoa_pods_package(
            name=pod_info.name, version=pod_info.version, location=new_location
        )
        if package:
            packages.append(package)

    return packages


def _generate_relations(pods_info: list[PodInfo], packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_name = {package.name: package for package in packages}

    for pod_info in pods_info:
        package = packages_by_name.get(pod_info.name)
        if not package:
            continue

        relationships.extend(
            Relationship(
                from_=package_dependency.id_,
                to_=package.id_,
                type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
            )
            for dependency in pod_info.dependencies
            if (package_dependency := packages_by_name.get(dependency))
        )

    return relationships


def _extract_pod_info(pod: ParsedValue) -> tuple[str, str]:
    if isinstance(pod, str | IndexedDict):
        pod_blob = pod if isinstance(pod, str) else next(iter(pod))
        pod_name = pod_blob.split(" ")[0]
        pod_version = pod_blob.split(" ")[1].strip("()")
        return pod_name, pod_version
    return "", ""


def _extract_pod_dependencies(pod: ParsedValue) -> list[str]:
    if not isinstance(pod, IndexedDict):
        return []

    dependencies: list[str] = []
    for value in pod.values():
        if not isinstance(value, IndexedList):
            continue
        dependencies.extend(dep.split(" ")[0] for dep in value if isinstance(dep, str))
    return dependencies



from collections.abc import Generator
from typing import TextIO

from labels.model.file import Location
from labels.model.resolver import FileReader, Resolver


class FakeResolver(Resolver):
    def __init__(
        self,
        *,
        file_content: TextIO | None = None,
        glob_results: list[Location] | None = None,
    ) -> None:
        self._file_content = file_content
        self._glob_results = glob_results or []

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str = "r",
    ) -> TextIO | None:
        return self._file_content

    def has_path(self, path: str) -> bool:
        return False

    def files_by_path(self, *paths: str) -> list[Location]:
        return []

    def files_by_glob(self, *patterns: str) -> list[Location]:
        return self._glob_results

    def relative_file_path(self, _: Location, path: str) -> Location | None:
        return None

    def walk_file(self) -> Generator[str, None, None]:
        yield from ()



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dart.package_builder import new_pubspec_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


def parse_pubspec_yaml(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_yaml_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    packages = _collect_packages(file_content, reader)

    return packages, []


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    deps = file_content.get("dependencies")
    dev_deps = file_content.get("dev_dependencies")

    return [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue,
    *,
    is_dev: bool = False,
) -> list[Package]:
    packages: list[Package] = []
    if not isinstance(dependencies, IndexedDict):
        return packages

    for name, version in dependencies.items():
        if not name or not isinstance(version, str):
            continue

        new_location = get_enriched_location(
            reader.location,
            line=dependencies.get_key_position(name).start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.PUB,
        )

        package = new_pubspec_package(name, version, new_location)
        if package:
            packages.append(package)

    return packages



import re
from typing import NamedTuple

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProject
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location

QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r'[^"\']+'
LINE_COMMENT_RE = re.compile(rf"^.*{NL}//.*$")

CONFIG_TO_DEV_STATUS = {
    "testRuntimeOnly": True,
    "testCompileOnly": True,
    "testImplementation": True,
    "compileOnly": True,
    "testCompile": True,
    "compile": False,
    "runtime": False,
    "implementation": False,
    "api": False,
    "runtimeOnly": False,
}


class GradleDependency(NamedTuple):
    group: str
    name: str
    version: str
    config: str
    line: int | None = None


def parse_gradle(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    gradle_dependencies = _parse_dependencies(reader.read_closer.read())

    packages = _collect_packages(gradle_dependencies, reader.location)

    return packages, []


def _collect_packages(
    gradle_dependencies: list[GradleDependency], reader_location: Location
) -> list[Package]:
    packages: list[Package] = []

    for dependency in gradle_dependencies:
        name = dependency.name
        version = dependency.version

        is_dev = CONFIG_TO_DEV_STATUS.get(dependency.config)

        new_location = get_enriched_location(
            reader_location,
            line=dependency.line,
            is_dev=is_dev,
            package_manager=PackageManager.GRADLE,
        )

        ecosystem_data = JavaArchive(
            pom_project=JavaPomProject(
                group_id=dependency.group,
                name=name,
                artifact_id=name,
                version=version,
            ),
        )

        package_spec = JavaPackageSpec(
            simple_name=name.rsplit(":", 1)[-1],
            composed_name=name,
            version=version,
            location=new_location,
            ecosystem_data=ecosystem_data,
        )

        package = new_java_package(package_spec)
        if package:
            packages.append(package)

    return packages


def _parse_dependencies(file_content: str) -> list[GradleDependency]:
    configs = _extract_gradle_configs(file_content)
    maven_regexes = _build_regex_with_configs(configs)

    dependencies = _get_block_deps(file_content, maven_regexes)

    is_block_comment = False
    for line_no, raw_line in enumerate(file_content.splitlines(), start=1):
        clean_line, is_block_comment = _avoid_cmt(raw_line, is_block_cmt=is_block_comment)
        dependency = _parse_line_dependency_line(clean_line, line_no, maven_regexes)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _extract_gradle_configs(content: str) -> set[str]:
    config_pattern = re.compile(r"configurations\s*\{([^}]+)\}", re.DOTALL)
    custom_config_pattern = re.compile(r"\s*(\w+)\s*")
    configs = {
        "runtimeOnly",
        "api",
        "compile",
        "compileOnly",
        "implementation",
        "testRuntimeOnly",
        "testCompileOnly",
        "testImplementation",
        "runtime",
    }

    config_blocks: list[str] = config_pattern.findall(content)
    for block in config_blocks:
        custom_configs: list[str] = custom_config_pattern.findall(block)
        configs.update(custom_configs)

    return configs


def _build_regex_with_configs(configs: set[str]) -> dict[str, re.Pattern[str]]:
    config_pattern = "|".join(configs)

    return {
        "RE_GRADLE_A": re.compile(
            rf"^{NL}(?P<config_name>{config_pattern}){NL}[(]?{NL}"
            rf"group{NL}:{NL}{QUOTE}(?P<group>{TEXT}){QUOTE}{NL},"
            rf"{NL}name{NL}:{NL}{QUOTE}(?P<name>{TEXT}){QUOTE}{NL}"
            rf"(?:,{NL}version{NL}:{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL})"
            rf"?.*$",
        ),
        "RE_GRADLE_B": re.compile(
            rf"^.*{NL}(?P<config_name>{config_pattern}){NL}[(]?{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}",
        ),
        "RE_GRADLE_C": re.compile(
            rf"{NL}(?P<config_name>{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}{{({NL})version{NL}{{({NL})strictly{NL}\({NL}"
            rf"{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}{NL}}}",
            re.DOTALL,
        ),
        "BLOCK": re.compile(
            rf"{NL}(?P<config_name>{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}\{{(.*?version{NL}\{{.*?\}}){NL}\}}",
            re.DOTALL,
        ),
        "VERSION": re.compile(
            rf"version{NL}{{({NL})strictly{NL}\("
            rf"{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}",
            re.DOTALL,
        ),
    }


def _parse_line_dependency_line(
    line: str, line_no: int, regexes: dict[str, re.Pattern[str]]
) -> GradleDependency | None:
    for regex_key in ("RE_GRADLE_A", "RE_GRADLE_B"):
        if match := regexes[regex_key].match(line):
            if regex_key == "RE_GRADLE_A":
                group = match.group("group")
                product = f"{group}:{match.group('name')}"
                version = match.group("version") or ""
            else:
                statement = str(match.group("statement"))
                product, version = (
                    statement.rsplit(":", maxsplit=1)
                    if statement.count(":") >= 2
                    else (statement, "")
                )
                group = product.split(":")[0]

            # Assuming a wildcard in Maven if the version is not found can # result in issues.
            # https://gitlab.com/fluidattacks/universe/-/issues/5635
            if not version or re.match(r"\${.*}", version):
                return None

            return GradleDependency(
                group=group,
                name=product,
                version=version,
                line=line_no,
                config=match.group("config_name"),
            )

    return None


def _avoid_cmt(line: str, *, is_block_cmt: bool) -> tuple[str, bool]:
    if LINE_COMMENT_RE.match(line):
        line = line.split("//", 1)[0]
    if is_block_cmt:
        if "*/" in line:
            is_block_cmt = False
            line = line.split("*/", 1).pop()
        else:
            return "", is_block_cmt
    if "/*" in line:
        line_cmt_open = line.split("/*", 1)[0]
        if "*/" in line:
            line = line_cmt_open + line.split("*/", 1).pop()
        else:
            line = line_cmt_open
            is_block_cmt = True
    return line, is_block_cmt


def _get_block_deps(content: str, regexes: dict[str, re.Pattern[str]]) -> list[GradleDependency]:
    dependencies = []

    for block in regexes["BLOCK"].finditer(content):
        product = str(block.group("statement"))
        hit = regexes["VERSION"].search(block.group())
        if not hit:
            continue

        version = hit.group("version")

        config = block.group("config_name")

        line_no = _get_line_number(content, block.start())

        dependencies.append(
            GradleDependency(
                group=product.split(":")[0],
                name=product,
                version=version,
                line=line_no,
                config=config,
            ),
        )

    return dependencies


def _get_line_number(content: str, match_start: int) -> int:
    return content[:match_start].count("\n") + 2



from typing import NamedTuple

from labels.model.file import DependencyType, Location
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    get_direct_dependencies_with_fallback,
    is_version_compatible,
    normalize_npm_alias,
)
from labels.parsers.cataloger.javascript.package_builder import new_npm_package_from_lock
from labels.parsers.cataloger.utils import get_enriched_location, parse_integrity_to_health_metadata


class NpmContext(NamedTuple):
    lockfile_path: str
    parent_key: str
    key_map: dict[tuple[str, int], str]


class RelationshipContext(NamedTuple):
    packages: list[Package]
    dependency_map: DependencyMap
    location: Location
    key_map: dict[tuple[str, int], str]
    root_package: Package | None
    direct_dependencies: list[str]
    dependencies: IndexedDict[str, ParsedValue]


def _create_root_package(
    location: Location,
    file_content: IndexedDict[str, ParsedValue],
) -> Package | None:
    root_name = file_content.get("name")
    root_version = file_content.get("version")

    if not isinstance(root_name, str) or not isinstance(root_version, str):
        return None

    root_location = get_enriched_location(
        location,
        line=1,
        is_dev=False,
        is_transitive=False,
        package_manager=PackageManager.NPM,
    )
    update_root: dict[str, DependencyType] = {"dependency_type": DependencyType.ROOT}
    root_location = root_location.model_copy(deep=True, update=update_root)

    return new_npm_package_from_lock(
        location=root_location,
        name=root_name,
        value=file_content,
        lockfile_version=1,
    )


def parse_package_lock_v1(
    location: Location,
    file_content: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    dependencies = file_content.get("dependencies")
    if not isinstance(dependencies, IndexedDict):
        return [], []

    packages = []
    root_package = _create_root_package(location, file_content)
    if root_package:
        packages.append(root_package)

    direct_dependencies = _get_direct_dependencies_v1(location, dependencies)
    packages.extend(_collect_packages(location, dependencies, direct_dependencies))

    consolidated_packages = _consolidate_packages_v1(packages)

    # Get consolidated root package
    consolidated_root = None
    if root_package:
        for pkg in consolidated_packages:
            if pkg.name == root_package.name and pkg.version == root_package.version:
                consolidated_root = pkg
                break

    dependency_map = _build_dependency_map_v1(dependencies)
    key_map = _build_key_map_v1(dependencies)

    rel_context = RelationshipContext(
        packages=consolidated_packages,
        dependency_map=dependency_map,
        location=location,
        key_map=key_map,
        root_package=consolidated_root,
        direct_dependencies=direct_dependencies,
        dependencies=dependencies,
    )
    relationships = _build_relationships_v1(rel_context)

    return consolidated_packages, relationships


def _consolidate_packages_v1(packages: list[Package]) -> list[Package]:
    consolidated: dict[tuple[str, str], Package] = {}

    for pkg in packages:
        key = (pkg.name, pkg.version)
        if key in consolidated:
            existing_pkg = consolidated[key]
            for loc in pkg.locations:
                if loc not in existing_pkg.locations:
                    existing_pkg.locations.append(loc)
        else:
            consolidated[key] = pkg

    return list(consolidated.values())


def _get_direct_dependencies_fallback(dependencies: IndexedDict[str, ParsedValue]) -> list[str]:
    """Fallback: infer direct dependencies from lockfile (less accurate)."""
    transitives: set[str] = set()
    for details in dependencies.values():
        if not isinstance(details, IndexedDict):
            continue

        requires = details.get("requires")
        if not isinstance(requires, IndexedDict):
            continue

        transitives.update(requires.keys())

    return [dependency for dependency in dependencies if dependency not in transitives]


def _get_direct_dependencies_v1(
    location: Location,
    dependencies: IndexedDict[str, ParsedValue],
) -> list[str]:
    fallback = _get_direct_dependencies_fallback(dependencies)
    return get_direct_dependencies_with_fallback(location, fallback)


def _collect_packages(
    location: Location,
    dependencies: IndexedDict[str, ParsedValue],
    direct_dependencies: list[str],
) -> list[Package]:
    packages: list[Package] = []
    for dependency_key, dependency_value in dependencies.items():
        if not isinstance(dependency_value, IndexedDict):
            continue

        is_transitive = dependency_key not in direct_dependencies
        is_dev = dependency_value.get("dev") is True
        new_location = get_enriched_location(
            location,
            line=dependency_value.position.start.line,
            is_dev=is_dev,
            is_transitive=is_transitive,
            package_manager=PackageManager.NPM,
        )

        integrity_str = dependency_value.get("integrity")
        health_metadata = parse_integrity_to_health_metadata(
            str(integrity_str) if integrity_str else None
        )

        package = new_npm_package_from_lock(
            location=new_location,
            name=dependency_key,
            value=dependency_value,
            lockfile_version=1,
            health_metadata=health_metadata,
        )
        if package:
            packages.append(package)

        sub_dependencies = dependency_value.get("dependencies")
        if isinstance(sub_dependencies, IndexedDict):
            packages.extend(_collect_packages(location, sub_dependencies, direct_dependencies))

    return packages


def _build_dependency_map_v1(
    dependencies: IndexedDict[str, ParsedValue],
    result: DependencyMap | None = None,
    parent_context: str = "",
) -> DependencyMap:
    if result is None:
        result = {}

    for dependency_key, dependency_value in dependencies.items():
        if not isinstance(dependency_value, IndexedDict):
            continue

        requires_dict = dependency_value.get("requires")
        if isinstance(requires_dict, IndexedDict):
            full_key = f"{parent_context}>{dependency_key}" if parent_context else dependency_key
            result[full_key] = {k: str(v) for k, v in requires_dict.data.items()}

        sub_dependencies = dependency_value.get("dependencies")
        if isinstance(sub_dependencies, IndexedDict):
            full_key_for_recursion = (
                f"{parent_context}>{dependency_key}" if parent_context else dependency_key
            )
            _build_dependency_map_v1(sub_dependencies, result, full_key_for_recursion)

    return result


def _build_key_map_v1(
    dependencies: IndexedDict[str, ParsedValue],
    result: dict[tuple[str, int], str] | None = None,
    parent_context: str = "",
) -> dict[tuple[str, int], str]:
    if result is None:
        result = {}

    for dependency_key, dependency_value in dependencies.items():
        if not isinstance(dependency_value, IndexedDict):
            continue

        line = dependency_value.position.start.line
        full_key = f"{parent_context}>{dependency_key}" if parent_context else dependency_key
        result[(dependency_key, line)] = full_key

        version = dependency_value.get("version")
        if isinstance(version, str):
            normalized_name, _ = normalize_npm_alias(version)
            if normalized_name:
                result[(normalized_name, line)] = full_key

        sub_dependencies = dependency_value.get("dependencies")
        if isinstance(sub_dependencies, IndexedDict):
            _build_key_map_v1(sub_dependencies, result, full_key)

    return result


def _resolve_child_location_v1(
    parent_key: str,
    dependency_name: str,
    dep_locations: list[Location],
    key_map: dict[tuple[str, int], str],
) -> Location | None:
    """Resolve the correct child location for npm v1."""
    candidate_keys = [f"{parent_key}>{dependency_name}"]

    parts = parent_key.split(">")
    for i in range(len(parts) - 1, -1, -1):
        ancestor_key = ">".join(parts[:i])
        if ancestor_key:
            candidate_keys.append(f"{ancestor_key}>{dependency_name}")

    candidate_keys.append(dependency_name)

    for candidate_key in candidate_keys:
        for dep_location in dep_locations:
            line = dep_location.coordinates.line if dep_location.coordinates else None
            if line is None:
                continue

            actual_key = key_map.get((dependency_name, line))

            if actual_key == candidate_key:
                return dep_location

    return None


def _normalize_npm_alias_name(alias_name: str, dependencies: IndexedDict[str, ParsedValue]) -> str:
    """Normalize npm alias name to the actual package name."""
    dep_value = dependencies.get(alias_name)
    if isinstance(dep_value, IndexedDict):
        version = dep_value.get("version")
        if isinstance(version, str):
            normalized_name, _ = normalize_npm_alias(version)
            if normalized_name:
                return normalized_name
    return alias_name


def _add_root_dependency_relationship(  # noqa: PLR0913
    dep_name: str,
    ctx: RelationshipContext,
    packages_by_name: dict[str, list[Package]],
    root_location: Location,
    lockfile_path: str,
    relationships: list[Relationship],
) -> None:
    """Add relationship for a single direct dependency to root."""
    normalized_name = _normalize_npm_alias_name(dep_name, ctx.dependencies)

    for dep_pkg in packages_by_name.get(normalized_name, []):
        all_locs = [loc for loc in dep_pkg.locations if loc.path() == lockfile_path]
        for dep_loc in all_locs:
            line = dep_loc.coordinates.line if dep_loc.coordinates else None
            if line is None:
                continue

            dep_key = ctx.key_map.get((dep_name, line))
            if not dep_key:
                dep_key = ctx.key_map.get((normalized_name, line))

            if dep_key in (dep_name, normalized_name) and ctx.root_package:
                from_id = f"{dep_pkg.id_}@{dep_loc.location_id()}"
                to_id = f"{ctx.root_package.id_}@{root_location.location_id()}"
                relationships.append(
                    Relationship(
                        from_=from_id,
                        to_=to_id,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                )
                break


def _create_root_relationships(
    ctx: RelationshipContext,
    packages_by_name: dict[str, list[Package]],
    lockfile_path: str,
) -> list[Relationship]:
    """Create relationships from direct dependencies to root package."""
    relationships: list[Relationship] = []

    if not ctx.root_package:
        return relationships

    root_location = None
    for loc in ctx.root_package.locations:
        if loc.path() == lockfile_path:
            root_location = loc
            break

    if not root_location:
        return relationships

    for dep_name in ctx.direct_dependencies:
        _add_root_dependency_relationship(
            dep_name, ctx, packages_by_name, root_location, lockfile_path, relationships
        )

    return relationships


def _build_relationships_v1(ctx: RelationshipContext) -> list[Relationship]:
    """Build relationships for npm v1 using location resolution."""
    lockfile_path = ctx.location.path()
    relationships: list[Relationship] = []

    packages_by_name: dict[str, list[Package]] = {}
    for pkg in ctx.packages:
        packages_by_name.setdefault(pkg.name, []).append(pkg)

    relationships.extend(_create_root_relationships(ctx, packages_by_name, lockfile_path))

    for package in ctx.packages:
        package_lock_locations = [loc for loc in package.locations if loc.path() == lockfile_path]
        if not package_lock_locations:
            continue

        for package_location in package_lock_locations:
            line = package_location.coordinates.line if package_location.coordinates else None
            if line is None:
                continue

            dep_key = ctx.key_map.get((package.name, line))
            if not dep_key:
                continue

            dependencies = ctx.dependency_map.get(dep_key)
            if not dependencies:
                continue

            npm_ctx = NpmContext(lockfile_path, dep_key, ctx.key_map)
            relationships.extend(
                _create_npm_relationships(
                    package, package_location, dependencies, packages_by_name, npm_ctx
                )
            )

    return relationships


def _create_npm_relationships(
    package: Package,
    package_location: Location,
    dependencies: dict[str, str],
    packages_by_name: dict[str, list[Package]],
    ctx: NpmContext,
) -> list[Relationship]:
    """Create npm relationships with location resolution."""
    relationships: list[Relationship] = []

    for dependency_name, required_version in dependencies.items():
        search_name = dependency_name
        actual_version = required_version

        if isinstance(required_version, str):
            normalized_name, normalized_version = normalize_npm_alias(required_version)
            if normalized_name:
                search_name = normalized_name
                actual_version = normalized_version

        for dep_pkg in packages_by_name.get(search_name, []):
            if not is_version_compatible(dep_pkg.version, actual_version):
                continue

            all_locs = [loc for loc in dep_pkg.locations if loc.path() == ctx.lockfile_path]
            resolved_location = _resolve_child_location_v1(
                ctx.parent_key, dependency_name, all_locs, ctx.key_map
            )

            if resolved_location:
                from_id = f"{dep_pkg.id_}@{resolved_location.location_id()}"
                to_id = f"{package.id_}@{package_location.location_id()}"
                relationships.append(
                    Relationship(
                        from_=from_id,
                        to_=to_id,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                )

    return relationships



import logging
from datetime import UTC, datetime
from typing import TYPE_CHECKING

from spdx_tools.spdx.model.actor import Actor, ActorType
from spdx_tools.spdx.model.document import CreationInfo, Document
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.writer.write_anything import write_file

from labels.model.core import OutputFormat, SbomConfig
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.output.spdx.complete_file import add_empty_package
from labels.output.spdx.file_builder import add_packages_and_relationships
from labels.output.utils import get_document_namespace, set_namespace_version
from labels.utils.exceptions import SPDXValidationError

if TYPE_CHECKING:  # pragma: no cover
    from spdx_tools.spdx.validation.validation_message import ValidationMessage

LOGGER = logging.getLogger(__name__)

_FORMAT_EXTENSION_MAP = {
    OutputFormat.SPDX_XML: "xml",
    OutputFormat.SPDX_JSON: "json",
}


def validate_spdx_sbom(file_format: str, file_name: str, document: Document) -> None:
    validation_errors: list[ValidationMessage] = validate_full_spdx_document(document)

    if validation_errors:
        raise SPDXValidationError(validation_errors)

    LOGGER.info(
        "🆗 Valid SPDX %s format, generating output file at %s.%s",
        file_format.upper(),
        file_name,
        file_format,
    )


def format_spdx_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver | SyftSBOM,
) -> None:
    now_utc = datetime.now(UTC)
    namespace, _ = set_namespace_version(config=config, resolver=resolver)
    creation_info = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name=namespace,
        data_license="CC0-1.0",
        document_namespace=get_document_namespace(namespace),
        creators=[Actor(ActorType.TOOL, "Fluid-Labels", None)],
        created=now_utc,
    )

    document = Document(creation_info)

    if not packages and not relationships:
        add_empty_package(document)
    else:
        add_packages_and_relationships(document, packages, relationships)

    file_format = _FORMAT_EXTENSION_MAP[config.output_format]

    validate_spdx_sbom(file_format, config.output, document)

    write_file(document, f"{config.output}.{file_format}")
    LOGGER.info("✅ Output file successfully generated")



import sqlite3

from fluidattacks_core.semver.match_versions import match_vulnerable_versions

from labels.advisories.database import BaseDatabase
from labels.advisories.match_fixes import match_fixed_versions
from labels.advisories.utils import create_advisory_from_record
from labels.model.advisories import Advisory, AdvisoryRecord


class ImagesDatabase(BaseDatabase):
    def __init__(self) -> None:
        super().__init__(db_name="skims_sca_advisories_for_images.db")


DATABASE = ImagesDatabase()


def _normalize_platform_version_for_query(platform_version: str) -> str:
    return platform_version.split(".")[0]


def fetch_advisory_from_database(
    cursor: sqlite3.Cursor,
    ecosystem: str,
    platform_version: str,
    package_name: str,
) -> list[AdvisoryRecord]:
    normalized_platform_version = (
        _normalize_platform_version_for_query(platform_version)
        if ecosystem == "rpm"
        else platform_version
    )

    cursor.execute(
        """
        SELECT
            adv_id,
            source,
            vulnerable_version,
            severity_level,
            severity_v4,
            epss,
            details,
            percentile,
            cwe_ids,
            cve_finding,
            auto_approve,
            fixed_versions,
            kev_catalog,
            platform_version
        FROM advisories
        WHERE ecosystem = ? AND platform_version = ? AND package_name = ?;
        """,
        (ecosystem, normalized_platform_version, package_name),
    )
    records: list[AdvisoryRecord] = cursor.fetchall()
    return records


def get_package_advisories(
    ecosystem: str,
    package_name: str,
    version: str,
    platform_version: str,
    upstream_package: str | None = None,
) -> list[Advisory]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()

    return [
        adv
        for record in fetch_advisory_from_database(
            cursor,
            ecosystem,
            platform_version,
            package_name,
        )
        if (
            adv := create_advisory_from_record(
                record, ecosystem, package_name, version, upstream_package
            )
        )
    ]


def get_vulnerabilities(
    ecosystem: str,
    product: str,
    version: str,
    platform_version: str | None,
    upstream_info: tuple[str | None, str | None] | None = None,
) -> list[Advisory]:
    advisories_list = []
    if (
        product
        and version
        and platform_version
        and (
            advisories := get_package_advisories(
                ecosystem,
                product.lower(),
                version,
                platform_version,
            )
        )
    ):
        advisories_list.extend(
            [
                match_fixed_versions(version.lower(), advisor, None)
                for advisor in advisories
                if match_vulnerable_versions(version.lower(), advisor.vulnerable_version)
            ]
        )
    upstream_package, upstream_version = upstream_info or (None, None)
    if (
        upstream_package
        and upstream_version
        and platform_version
        and (
            origin_advisories := get_package_advisories(
                ecosystem,
                upstream_package.lower(),
                upstream_version,
                platform_version,
                upstream_package,
            )
        )
    ):
        advisories_list.extend(
            [
                match_fixed_versions(version.lower(), advisor, None)
                for advisor in origin_advisories
                if match_vulnerable_versions(version.lower(), advisor.vulnerable_version)
            ]
        )

    return advisories_list



from labels.model.ecosystem_data.java import (
    JavaArchive,
    JavaManifest,
    JavaPomProject,
    JavaPomProperties,
)
from labels.parsers.cataloger.java.utils.group_id_pkg_list import group_id_from_known_package_list


def group_id_from_java_metadata(pkg_name: str | None, metadata: JavaArchive | None) -> str | None:
    if not metadata or not pkg_name:
        return None

    # 1. Group ID from POM properties
    if group_id := _group_id_from_pom_properties(metadata.pom_properties):
        return group_id

    # 2. Group ID from POM project
    if group_id := _group_id_pom_project(metadata.pom_project):
        return group_id

    # 3. Group ID from known package list
    if group_id := group_id_from_known_package_list(pkg_name):
        return group_id

    # 4. Group ID from Java manifest
    if group_id := _group_id_from_java_manifest(metadata.manifest):
        return group_id

    return None


def _group_id_from_pom_properties(properties: JavaPomProperties | None) -> str:
    if not properties:
        return ""
    if properties.group_id:
        return _clean_group_id(properties.group_id)
    if properties.artifact_id and _looks_like_group_id(properties.artifact_id):
        return _clean_group_id(properties.artifact_id)
    return ""


def _group_id_pom_project(project: JavaPomProject | None) -> str:
    if not project:
        return ""
    if project.group_id:
        return _clean_group_id(project.group_id)
    if project.artifact_id and _looks_like_group_id(project.artifact_id):
        return _clean_group_id(project.artifact_id)
    if project.parent:
        if project.parent.group_id:
            return _clean_group_id(project.parent.group_id)
        if _looks_like_group_id(project.parent.artifact_id):
            return _clean_group_id(project.parent.artifact_id)
    return ""


def _group_id_from_java_manifest(manifest: JavaManifest | None) -> str:
    if not manifest or not manifest.main:
        return ""

    primary_fields = [
        "Group-Id",
        "Bundle-SymbolicName",
        "Extension-Name",
        "Specification-Vendor",
        "Implementation-Vendor",
        "Implementation-Vendor-Id",
        "Implementation-Title",
        "Bundle-Activator",
    ]

    group_ids = _get_group_ids_from_manifest_fields(manifest, primary_fields)
    if group_ids:
        return group_ids[0]

    secondary_fields = ["Automatic-Module-Name", "Main-Class", "Package"]
    group_ids = _get_group_ids_from_manifest_fields(manifest, secondary_fields)
    return group_ids[0] if group_ids else ""


def _get_group_ids_from_manifest_fields(manifest: JavaManifest, fields: list[str]) -> list[str]:
    group_ids = []
    for field in fields:
        if (value := manifest.main.get(field, "")) and _starts_with_top_level_domain(value):
            group_ids.append(_clean_group_id(value))
        if manifest.sections:
            section_values = [
                _clean_group_id(value)
                for section in manifest.sections
                if (value := section.get(field, "")) and _starts_with_top_level_domain(value)
            ]
            group_ids.extend(section_values)
    return group_ids


def _starts_with_top_level_domain(value: str) -> bool:
    domains = ["com", "org", "net", "io", "be"]
    return any(value.startswith(domain + ".") for domain in domains)


def _clean_group_id(group_id: str) -> str:
    return _remove_osgi_directives(group_id).strip()


def _remove_osgi_directives(group_id: str) -> str:
    return group_id.split(";")[0]


def _looks_like_group_id(group_id: str) -> bool:
    return "." in group_id



import re

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.utils import get_enriched_location

GRADLE_DISTRIBUTION = re.compile("^distributionUrl=.+gradle-(?P<gradle_version>[^-]+)-.+")


def parse_gradle_properties(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = _collect_packages(reader.read_closer.read(), reader.location)

    return packages, []


def _collect_packages(file_content: str, location: Location) -> list[Package]:
    packages: list[Package] = []

    for line_no, raw_line in enumerate(file_content.splitlines(), start=1):
        line = raw_line.strip()

        matched_gradle_distribution = GRADLE_DISTRIBUTION.match(line)
        if not matched_gradle_distribution:
            continue

        version = matched_gradle_distribution.group("gradle_version")

        new_location = get_enriched_location(
            location, line=line_no, package_manager=PackageManager.GRADLE
        )

        package_spec = JavaPackageSpec(
            simple_name="gradle",
            version=version,
            location=new_location,
        )

        package = new_java_package(package_spec)

        if package:  # pragma: no branch
            packages.append(package)

    return packages



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Parser, Request
from labels.parsers.cataloger.python.parse_pipfile_deps import parse_pipfile_deps
from labels.parsers.cataloger.python.parse_pipfile_lock import parse_pipfile_lock_deps
from labels.parsers.cataloger.python.parse_poetry_lock import parse_poetry_lock
from labels.parsers.cataloger.python.parse_pyproject_toml import parse_pyproject_toml
from labels.parsers.cataloger.python.parse_requirements import parse_requirements_txt
from labels.parsers.cataloger.python.parse_uv_lock import parse_uv_lock
from labels.parsers.cataloger.python.parse_wheel_egg import parse_wheel_or_egg


def _get_parser_config(value: str) -> tuple[str, Parser, str] | None:
    patterns = [
        (
            ("*.txt", "*/*.txt", "*requirements.in", "requirements.in", "*/requirements.in"),
            parse_requirements_txt,
            "python-requirements-cataloger",
        ),
        (
            ("*poetry.lock", "poetry.lock", "*/poetry.lock"),
            parse_poetry_lock,
            "python-poetry-lock-cataloger",
        ),
        (("*uv.lock", "uv.lock", "*/uv.lock"), parse_uv_lock, "python-uv-lock-cataloger"),
        (
            (
                "**/*.egg-info",
                "**/*dist-info/METADATA",
                "**/*egg-info/PKG-INFO",
                "**/*DIST-INFO/METADATA",
                "**/*EGG-INFO/PKG-INFO",
            ),
            parse_wheel_or_egg,
            "python-installed-package-cataloger",
        ),
        (
            ("**/Pipfile.lock", "Pipfile.lock"),
            parse_pipfile_lock_deps,
            "python-pipfile-lock-cataloger",
        ),
        (("**/Pipfile", "Pipfile"), parse_pipfile_deps, "python-pipfile-package-cataloger"),
        (
            ("**/pyproject.toml", "pyproject.toml"),
            parse_pyproject_toml,
            "python-pyproject-toml-cataloger",
        ),
    ]

    for pattern_list, parser_func, parser_name in patterns:
        if any(fnmatch(value, pattern) for pattern in pattern_list):
            return value, parser_func, parser_name
    return None


def on_next_python(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                config = _get_parser_config(value)
                if config:
                    real_path, parser_func, parser_name = config
                    observer.on_next(
                        Request(
                            real_path=real_path,
                            parser=parser_func,
                            parser_name=parser_name,
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import re

from bs4 import BeautifulSoup, Tag

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location

SCRIPT_DEP = re.compile(
    r"(?P<name>[^\s\/]*)(?P<separator>[-@\/])"
    r"(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))",
)

NPM_CDNS = (
    "https://cdn.jsdelivr.net/npm/",
    "https://unpkg.com/",
    "https://cdn.skypack.dev/",
    "https://cdn.esm.sh/",
    "https://code.jquery.com/",
)


def parse_html_scripts(
    _resolver: Resolver | None, _env: Environment | None, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    html = _safe_parse_html(reader)
    if not html:
        return [], []

    packages = _collect_packages(html, reader)

    return packages, []


def _collect_packages(html: BeautifulSoup, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []
    html_scripts: list[Tag] = html.find_all("script")
    for script in html_scripts:
        src_attribute = str(script.get("src"))
        if not (src_attribute and src_attribute.endswith(".js")):
            continue

        if not src_attribute.startswith(NPM_CDNS):
            continue

        name_version = _parse_name_version_from_src(src_attribute)
        if not name_version:
            continue

        name, version = name_version

        new_location = get_enriched_location(reader.location, line=script.sourceline)

        package = new_simple_npm_package(new_location, name, version)
        if package:
            packages.append(package)

    return packages


def _safe_parse_html(reader: LocationReadCloser) -> BeautifulSoup | None:
    try:
        return BeautifulSoup(reader.read_closer, features="html.parser")
    except (UnicodeError, AssertionError):
        return None


def _parse_name_version_from_src(src: str) -> tuple[str, str] | None:
    matched = SCRIPT_DEP.search(src)
    if not matched:
        return None

    return matched.group("name"), matched.group("version")



import logging
import re
from contextlib import suppress

import tree_sitter_yaml
from tree_sitter import Language as TLanguage
from tree_sitter import Node, Parser

from labels.model.indexables import FileCoordinate, IndexedDict, IndexedList, ParsedValue, Position
from labels.utils.exceptions import (
    UnexpectedChildrenLengthError,
    UnexpectedNodeError,
    UnexpectedNodeTypeError,
)

LOGGER = logging.getLogger(__name__)


def _generate_position(node: Node) -> Position:
    return Position(
        start=FileCoordinate(line=node.start_point[0] + 1, column=node.start_point[1] + 1),
        end=FileCoordinate(line=node.end_point[0] + 1, column=node.end_point[1] + 1),
    )


def _validate_block_mapping_children(
    children: tuple[Node | None, ...], node: Node, *, was_padded: bool
) -> bool:
    if not was_padded:
        return True
    if len(children) < 2 or children[1] is None or children[1].type != ":":
        LOGGER.exception(
            "Unexpected child node structure. Expected 3 children, but got %d",
            len(children),
            exc_info=UnexpectedChildrenLengthError(node, 3),
            extra={"node": node, "children": children},
        )
        return False
    return True


def _get_mapping_children(
    child: Node, node: Node
) -> tuple[Node | None, Node | None, Node | None] | None:
    children: tuple[Node | None, ...] = tuple(x for x in child.children if x.type != "comment")
    was_padded = len(children) != 3
    if was_padded:
        children = (*children, None)
    if not _validate_block_mapping_children(children, node, was_padded=was_padded):
        return None
    key_up_node, _, value_up_node = children
    return key_up_node, _, value_up_node


def _process_block_mapping_pair(
    child: Node, node: Node, data: IndexedDict[str, ParsedValue]
) -> None:
    result = _get_mapping_children(child, node)
    if result is None:
        return
    key_up_node, _, value_up_node = result
    if key_up_node is None:
        return
    key_node, key_value = handle_node(key_up_node)
    if not isinstance(key_value, str):
        return
    if value_up_node is None:
        data[(key_value, _generate_position(key_node))] = (
            None,
            _generate_position(key_node),
        )
        return
    value_node, value_value = handle_node(value_up_node)
    data[(key_value, _generate_position(key_node))] = (
        value_value,
        _generate_position(value_node),
    )


def _handle_block_mapping_node(
    node: Node,
) -> tuple[Node, IndexedDict[str, ParsedValue]]:
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for child in node.children:
        if child.type == "block_mapping_pair":
            _process_block_mapping_pair(child, node, data)
    return node, data


def _handle_flow_mapping_node(node: Node) -> tuple[Node, IndexedDict[str, ParsedValue]]:
    pair_nodes = [x for x in node.children if x.type == "flow_pair"]
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for pair_node in pair_nodes:
        if len(pair_node.children) != 3:
            raise UnexpectedNodeError(pair_node)
        key_up_node, _, value_up_node = pair_node.children
        key_node, key_value = handle_node(key_up_node)
        value_node, value_value = handle_node(value_up_node)
        if not isinstance(key_value, str):
            continue
        data[(key_value, _generate_position(key_node))] = (
            value_value,
            _generate_position(value_node),
        )
    return node, data


def _handle_boolean_scalar_node(node: Node) -> tuple[Node, bool]:
    node_string_value = node.text.decode("utf-8").lower() if node.text else "false"
    return node, node_string_value == "true"


def _handle_block_sequence_node(
    node: Node,
) -> tuple[Node, IndexedList[ParsedValue]]:
    data: IndexedList[ParsedValue] = IndexedList(node)
    for child in (x for x in node.children if x.type != "comment"):
        if child.type != "block_sequence_item":
            raise UnexpectedNodeTypeError(child.type, "block_sequence_item")
        resolved_item = handle_node(child)
        data.append((resolved_item[1], _generate_position(resolved_item[0])))
    return node, data


def _handle_block_sequence_item(node: Node) -> tuple[Node, ParsedValue]:
    if len(node.children) != 2 or node.children[0].type != "-":
        raise UnexpectedNodeTypeError(node)
    return handle_node(node.children[1])


def _handle_integer_scalar_node(node: Node) -> tuple[Node, int]:
    decode_str = node.text.decode("utf-8") if node.text else ""
    with suppress(ValueError):
        return node, int(decode_str)
    with suppress(ValueError):
        return node, int(decode_str, 16)

    error_msg = f"Invalid integer value: {decode_str}"
    raise ValueError(error_msg)


def _handle_flow_sequence_node(
    node: Node,
) -> tuple[Node, IndexedList[ParsedValue]]:
    data: IndexedList[ParsedValue] = IndexedList(node)
    for child in [x for x in node.children if x.type not in ("[", "]", ",")]:
        if child.type != "flow_node":
            raise UnexpectedNodeTypeError(child.type, "flow_node")
        resolved_node, resolved_item = handle_node(child)
        data.append((resolved_item, _generate_position(resolved_node)))
    return node, data


def _handle_float_scalar_node(node: Node) -> tuple[Node, float]:
    decoded_str = node.text.decode("utf-8") if node.text else ""
    with suppress(ValueError):
        return node, float(decoded_str)

    with suppress(ValueError):
        return node, float(decoded_str.replace(".", "").lower())

    error_msg = f"Invalid float value: {decoded_str}"
    raise ValueError(error_msg)


def _handle_block_scalar(node: Node) -> tuple[Node, str]:
    decoded_str = node.text.decode("utf-8") if node.text else ""
    value = ""
    if match := re.search(r"^>(\d+)", decoded_str):
        indent_spaces = int(match.group(1))
        decoded_str = re.sub(r"^>\d+", "", decoded_str).strip()
    else:
        indent_spaces = 1  # Default to no indent removal if no marker is found

    if decoded_str.startswith(">"):
        value = (" " * indent_spaces).join(decoded_str.lstrip(">+-").strip().split())
    elif decoded_str.startswith("|"):
        normalized_str = decoded_str.replace("\xa0", " ").lstrip("|+-")
        value = "\n".join(line.strip() for line in normalized_str.split("\n"))
        value = value.replace("\n", "", 1)
    return node, value


def _handle_block_node(node: Node) -> tuple[Node, ParsedValue]:
    if len(node.children) > 1:
        raise UnexpectedChildrenLengthError(node.type, 1)
    return handle_node(node.children[0])


def _handle_flow_node(node: Node) -> tuple[Node, ParsedValue]:
    if len(node.children) > 1:
        raise UnexpectedChildrenLengthError(node.type, 1)
    return handle_node(node.children[0])


def _handle_plain_scalar(node: Node) -> tuple[Node, ParsedValue]:
    if len(node.children) > 1:
        raise UnexpectedChildrenLengthError(node.type, 1)
    return handle_node(node.children[0])


def _handle_string_scalar(node: Node) -> tuple[Node, str]:
    return node, node.text.decode("utf-8") if node.text else ""


def _handle_quote_scalar(node: Node) -> tuple[Node, str]:
    return node, node.text.decode("utf-8").strip("'\"") if node.text else ""


def _handle_null_scalar(node: Node) -> tuple[Node, None]:
    return node, None


_NODE_HANDLERS = {
    "block_node": _handle_block_node,
    "block_mapping": _handle_block_mapping_node,
    "flow_node": _handle_flow_node,
    "plain_scalar": _handle_plain_scalar,
    "string_scalar": _handle_string_scalar,
    "single_quote_scalar": _handle_quote_scalar,
    "double_quote_scalar": _handle_quote_scalar,
    "float_scalar": _handle_float_scalar_node,
    "flow_mapping": _handle_flow_mapping_node,
    "boolean_scalar": _handle_boolean_scalar_node,
    "null_scalar": _handle_null_scalar,
    "integer_scalar": _handle_integer_scalar_node,
    "block_sequence_item": _handle_block_sequence_item,
    "block_sequence": _handle_block_sequence_node,
    "flow_sequence": _handle_flow_sequence_node,
    "block_scalar": _handle_block_scalar,
}


def handle_node(node: Node) -> tuple[Node, ParsedValue]:
    handler = _NODE_HANDLERS.get(node.type)
    if handler is None:
        raise UnexpectedNodeError(node.type)

    try:
        return handler(node)
    except Exception as e:
        if isinstance(
            e,
            UnexpectedNodeError
            | ValueError
            | UnexpectedNodeTypeError
            | UnexpectedChildrenLengthError,
        ):
            raise
        raise UnexpectedNodeError(node.type) from e


def parse_yaml_with_tree_sitter(content: str) -> ParsedValue | None:
    parser_language = TLanguage(tree_sitter_yaml.language())
    parser = Parser(parser_language)
    result = parser.parse(content.encode("utf-8"))
    documents = [x for x in result.root_node.children if x.type == "document"]
    if len(documents) != 1:
        return None
    block_node = next((x for x in documents[0].children if x.type == "block_node"), None)
    if not block_node:
        return None
    try:
        _, value = handle_node(block_node)
    except (
        UnexpectedNodeError,
        ValueError,
        UnexpectedNodeTypeError,
        UnexpectedChildrenLengthError,
    ):
        LOGGER.exception(
            "Failed to handle block node of type '%s' while parsing YAML content",
            block_node.type,
            extra={"extra": {"node_type": block_node.type, "yaml": content}},
        )
        return None
    return value



def format_ruby_version_constraints(constraints: list[str]) -> str:
    if not constraints:
        return ""

    cleaned = [c.replace(" ", "") for c in constraints]
    first = cleaned[0]

    if "~>" not in first:
        return first if len(cleaned) == 1 else " ".join(cleaned)

    if len(cleaned) == 1:
        return first.replace("~>", "^" if first.count(".") <= 1 else "~")

    pessimistic_version = _format_pessimistic_with_bound(first, cleaned[1])

    return pessimistic_version if pessimistic_version else " ".join(cleaned)


def _format_pessimistic_with_bound(pessimistic: str, bound: str) -> str:
    if ">=" in bound:
        return bound.replace(">=", "^")
    if "!=" in bound:
        first_ver = pessimistic.replace("~>", "")
        second_ver = bound.replace("!=", "")
        return f">={first_ver} <{second_ver} || >{second_ver}"
    return ""



from labels.model.ecosystem_data.base import EcosystemDataModel


class DpkgDBEntry(EcosystemDataModel):
    package: str | None = None
    source: str | None = None
    version: str | None = None
    source_version: str | None = None
    architecture: str | None = None
    maintainer: str | None = None
    provides: list[str] | None = None
    dependencies: list[str] | None = None
    pre_dependencies: list[str] | None = None



import re
from typing import cast

import packaging.version
import pefile

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package

SPACE_REGEX = re.compile(r"[\s]+")
NUMBER_REGEX = re.compile(r"\d")
VERSION_PUNCTUATION_REGEX = re.compile(r"[.,]+")


def parse_dotnet_portable_executable(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    pe_representation = _safe_parse_version_pe_file(reader)
    if not pe_representation:
        return [], []

    packages = _collect_packages(pe_representation, reader.location)

    return packages, []


def _safe_parse_version_pe_file(reader: LocationReadCloser) -> pefile.PE | None:
    if not reader.location.coordinates:
        return None

    try:
        pe_representation = pefile.PE(reader.location.coordinates.real_path, fast_load=False)
    except pefile.PEFormatError:  # type: ignore[misc]
        return None

    return pe_representation


def _collect_packages(pe_representation: pefile.PE, location: Location) -> list[Package]:
    packages: list[Package] = []

    version_resource = _parse_version_resource(pe_representation)
    if not version_resource:
        return packages

    name = _find_name(version_resource)
    version = _find_version(version_resource)

    dotnet_package = new_dotnet_package(name, version, location)
    if dotnet_package:
        packages.append(dotnet_package)

    return packages


def _parse_version_resource(portable: pefile.PE) -> dict[str, str] | None:
    if not hasattr(portable, "VS_VERSIONINFO"):
        return None

    for idx, _ in enumerate(portable.VS_VERSIONINFO):
        if not hasattr(portable, "FileInfo") or len(portable.FileInfo) <= idx:
            continue

        parsed_file_info = _process_file_info(portable.FileInfo[idx])
        if parsed_file_info:
            return parsed_file_info

    return None


def _process_file_info(
    file_info: list["pefile._StringFileInfo"],
) -> dict[str, str] | None:
    for entry in file_info:
        if not hasattr(entry, "StringTable"):
            continue

        parsed_file_info = _process_string_table(entry.StringTable)
        if parsed_file_info:
            return parsed_file_info

    return None


def _process_string_table(
    string_table: list["pefile._StringTable"],
) -> dict[str, str]:
    string_table_dict: dict[str, str] = {}

    for string_table_entry in string_table:
        string_table_dict["LangID"] = string_table_entry.LangID.decode("utf-8")
        for key, value in string_table_entry.entries.items():
            string_table_dict[key.decode("utf-8")] = value.decode("utf-8")

    return string_table_dict


def _find_name(version_resources: dict[str, str]) -> str:
    name_fields = [
        "ProductName",
        "FileDescription",
        "InternalName",
        "OriginalFilename",
    ]

    if _is_microsoft(version_resources):
        name_fields = [
            "FileDescription",
            "InternalName",
            "OriginalFilename",
            "ProductName",
        ]

    for field in name_fields:
        value = _space_normalize(version_resources.get(field, ""))
        if value:
            return value

    return ""


def _is_microsoft(version_resources: dict[str, str]) -> bool:
    company_name = version_resources.get("CompanyName", "").lower()
    product_name = version_resources.get("ProductName", "").lower()

    return "microsoft" in company_name or "microsoft" in product_name


def _space_normalize(value: str) -> str:
    value = value.strip()

    if value == "":
        return ""

    value = value.encode("utf-8", "replace").decode("utf-8")

    value = SPACE_REGEX.sub(" ", value)

    value = re.sub(r"[\x00-\x1f]", "", value)

    value = SPACE_REGEX.sub(" ", value)

    return value.strip()


def _find_version(version_resources: dict[str, str]) -> str:
    raw_product_version = version_resources.get("ProductVersion", "")
    raw_file_version = version_resources.get("FileVersion", "")

    product_version = _extract_version(raw_product_version)
    file_version = _extract_version(raw_file_version)

    if not product_version and not file_version:
        return ""

    if not product_version:
        return file_version

    if not file_version:
        return product_version

    semantic_choice = _keep_greater_semantic_version(product_version, file_version)
    if semantic_choice:
        return semantic_choice

    return _choose_by_detail_and_numbers(product_version, file_version)


def _choose_by_detail_and_numbers(product_version: str, file_version: str) -> str:
    product_detail = _punctuation_count(product_version)
    file_detail = _punctuation_count(file_version)

    has_num_product = _contains_number(product_version)
    has_num_file = _contains_number(file_version)

    if has_num_file and file_detail > 0 and (not has_num_product or file_detail > product_detail):
        return file_version

    if has_num_product:
        return product_version
    if has_num_file:
        return file_version

    return product_version or file_version


def _keep_greater_semantic_version(product_version: str, file_version: str) -> str:
    try:
        semantic_product_version = packaging.version.parse(product_version)
    except ValueError:
        return ""

    try:
        semantic_file_version = packaging.version.parse(file_version)
    except ValueError:
        return product_version

    if semantic_product_version == semantic_file_version:
        return ""

    if semantic_file_version > semantic_product_version:
        return file_version

    return product_version


def _punctuation_count(string: str) -> int:
    return len(cast("list[str]", VERSION_PUNCTUATION_REGEX.findall(string)))


def _extract_version(version: str) -> str:
    version = version.strip()

    result = ""

    for index, character in enumerate(version.split()):
        if _contains_number(result) and not _contains_number(character):
            return result

        if index == 0:
            result = character
        else:
            result += " " + character

    return result


def _contains_number(string: str) -> bool:
    return any(character.isdigit() for character in string)



import logging
import os
import stat
import zipfile
from collections.abc import Callable
from contextlib import suppress
from fnmatch import fnmatch
from pathlib import Path
from zipfile import BadZipFile, ZipFile, ZipInfo

LOGGER = logging.getLogger(__name__)


def normalize_zip_entry_name(entry: str, *, case_insensitive: bool) -> str:
    if case_insensitive:
        entry = entry.lower()
    if not entry.startswith("/"):
        entry = "/" + entry

    return entry


def new_normalize_zip_entry_name(entry: str, *, case_sensitive: bool) -> str:
    if not case_sensitive:
        entry = entry.lower()
    if not entry.startswith("/"):
        entry = "/" + entry

    return entry


def new_zip_file_manifest(archive_path: str) -> list[ZipInfo]:
    try:
        with ZipFile(archive_path, "r") as myzip:
            return myzip.infolist()
    except BadZipFile:
        return []


def zip_glob_match(
    manifest: list[ZipInfo],
    *,
    case_sensitive: bool,
    patterns: tuple[str, ...],
) -> list[str]:
    result = []

    for pattern in patterns:
        for entry in manifest:
            normalized_entry = normalize_zip_entry_name(
                entry.filename,
                case_insensitive=case_sensitive,
            )
            if entry.filename.endswith(pattern):
                result.append(entry.filename)
            lower_pattern = pattern.lower() if case_sensitive else pattern
            if fnmatch(normalized_entry, lower_pattern):
                result.append(entry.filename)
    return result


def new_zip_glob_match(
    manifest: list[ZipInfo],
    patterns: tuple[str, ...],
    *,
    case_sensitive: bool,
) -> list[str]:
    matched: set[str] = set()

    for pattern in patterns:
        for entry in manifest:
            normalized_entry = new_normalize_zip_entry_name(
                entry.filename,
                case_sensitive=case_sensitive,
            )
            lower_pattern = pattern.lower() if not case_sensitive else pattern
            if fnmatch(normalized_entry, lower_pattern):
                matched.add(entry.filename)
    return list(matched)


def traverse_files_in_zip(
    archive_path: str,
    visitor: Callable[[zipfile.ZipInfo], None],
    *paths: str,
) -> None:
    """Traverse files in a zip file applying a visitor function to each file."""
    with zipfile.ZipFile(archive_path, "r") as zip_reader:
        for path in paths:
            try:
                visitor(zip_reader.getinfo(path))
            except KeyError:
                LOGGER.exception("Unable to find file: %s", path)


def contents_from_zip(archive_path: str, *paths: str) -> dict[str, str]:
    """Extract specified files from a zip archive and return their contents."""
    results: dict[str, str] = {}

    if not paths:
        return results

    def visitor(file: zipfile.ZipInfo) -> None:
        """Visitor function to read the contents of a file in the zip."""
        if file.is_dir():
            LOGGER.error("Unable to extract directories, only files: %s", file.filename)
            return
        with zipfile.ZipFile(archive_path, "r") as zip_reader, zip_reader.open(file) as file_data:
            content = file_data.read()
            with suppress(UnicodeDecodeError):
                results[file.filename] = content.decode(
                    "utf-8",
                )

    traverse_files_in_zip(archive_path, visitor, *paths)
    return results


def safe_extract(apk_file: zipfile.ZipFile, destination: str) -> None:
    for file_info in apk_file.infolist():
        file_name = file_info.filename
        if Path(file_name).is_absolute() or file_name.startswith(("..", "./")):
            continue

        target_path = Path(destination, file_name)

        if not _is_safe_path(destination, str(target_path)):
            continue

        if (file_info.external_attr >> 16) & stat.S_IFLNK:
            continue

        try:
            apk_file.extract(file_name, destination)
        except Exception:
            LOGGER.exception("Error extracting %s", file_name)


def _is_safe_path(base_path: str, target_path: str) -> bool:
    base_path = os.path.normpath(base_path)
    target_path = os.path.normpath(target_path)
    return os.path.commonpath([base_path]) == os.path.commonpath([base_path, target_path])



from labels.model.ecosystem_data.base import EcosystemDataModel


class RpmDBEntry(EcosystemDataModel):
    name: str
    version: str
    epoch: int | None
    arch: str
    release: str
    source_rpm: str



from os import environ
from typing import Literal


def guess_environment() -> Literal["development", "production"]:
    return (
        "production"
        if environ.get("AWS_BATCH_JOB_ID") or environ.get("CI_COMMIT_REF_NAME") == "trunk"
        else "development"
    )



from decimal import Decimal

from cyclonedx.model import HashType, Property, XsUri
from cyclonedx.model.vulnerability import (
    BomTarget,
    BomTargetVersionRange,
    Vulnerability,
    VulnerabilityAdvisory,
    VulnerabilityRating,
    VulnerabilitySeverity,
)

from labels.model.file import Location
from labels.model.metadata import HealthMetadata
from labels.model.package import Package

NAMESPACE = "fluid-attacks"


def add_vulnerabilities(package: Package) -> list[Vulnerability]:
    vulnerabilities = []
    if package.advisories:
        for advisory in package.advisories:
            rating = VulnerabilityRating(
                severity=VulnerabilitySeverity(
                    advisory.severity_level.lower() if advisory.severity_level else None,
                ),
                score=Decimal(str(advisory.percentile)),
            )

            advisory_obj = (
                [VulnerabilityAdvisory(url=XsUri(advisory.source))] if advisory.source else None
            )

            bom_target = BomTarget(
                ref=f"{package.name}@{package.version}",
                versions=[
                    BomTargetVersionRange(
                        version=package.version if not advisory.vulnerable_version else None,
                        range=advisory.vulnerable_version,
                    ),
                ],
            )

            additional_properties = [
                Property(name=f"{NAMESPACE}:advisory:epss", value=str(advisory.epss)),
                Property(name=f"{NAMESPACE}:advisory:kev_catalog", value=str(advisory.kev_catalog)),
            ]

            vulnerability = Vulnerability(
                bom_ref=f"{package.name}@{advisory.id}",
                id=advisory.id,
                description=advisory.details,
                advisories=advisory_obj,
                ratings=[rating],
                affects=[bom_target],
                properties=additional_properties,
            )

            vulnerabilities.append(vulnerability)

    return vulnerabilities


def add_locations_properties(locations: list[Location]) -> list[Property]:
    properties = []
    for idx, location in enumerate(locations):
        path = location.path()

        line = (
            location.coordinates.line
            if location.coordinates and location.coordinates.line
            else None
        )
        layer = (
            location.coordinates.file_system_id
            if location.coordinates and location.coordinates.file_system_id
            else None
        )

        properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:path", value=path))

        if line:
            properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:line", value=str(line)))

        if layer:
            properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:layer", value=layer))

    return properties


def add_language_property(package_language: str) -> Property:
    return Property(name=f"{NAMESPACE}:language", value=package_language)


def add_integrity(health_metadata: HealthMetadata) -> list[HashType] | None:
    integrity = health_metadata.artifact.integrity if health_metadata.artifact else None
    if integrity and integrity.algorithm and integrity.value:
        return [HashType.from_hashlib_alg(integrity.algorithm, integrity.value)]
    return None


def add_component_properties(package: Package) -> list[Property]:
    properties = []

    properties.append(add_language_property(package.language))
    properties.extend(add_locations_properties(package.locations))

    return properties



from typing import cast

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_composer
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter

EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def parse_installed_json(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, (IndexedDict, IndexedList)):  # pragma: no cover
        return [], []

    packages = _extract_packages(file_content, reader.location)
    relationships = _extract_relationships(packages, file_content)

    return packages, relationships


def _extract_packages(
    file_content: IndexedDict[str, ParsedValue] | IndexedList[ParsedValue],
    location: Location,
) -> list[Package]:
    packages = []
    dev_packages = EMPTY_LIST

    if (
        isinstance(file_content, IndexedDict)
        and "dev-package-names" in file_content
        and isinstance(file_content["dev-package-names"], IndexedList)
    ):
        dev_packages = file_content["dev-package-names"]

    packages_list = (
        file_content["packages"]
        if isinstance(file_content, IndexedDict) and "packages" in file_content
        else file_content
    )

    for package in packages_list if isinstance(packages_list, IndexedList) else EMPTY_LIST:
        if not isinstance(package, IndexedDict):
            continue

        name = cast("str", package.get("name"))
        version = cast("str", package.get("version"))

        if not name:
            continue

        is_dev = _is_dev_package(package, dev_packages)
        new_location = get_enriched_location(
            location,
            line=package.get_key_position("name").start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.COMPOSER,
        )

        pkg_item = new_package_from_composer(name=name, version=version, location=new_location)
        if pkg_item:
            packages.append(pkg_item)

    return packages


def _extract_relationships(
    packages: list[Package], file_content: IndexedDict[str, ParsedValue] | IndexedList[ParsedValue]
) -> list[Relationship]:
    relationships = []
    packages_list = (
        file_content["packages"]
        if isinstance(file_content, IndexedDict) and "packages" in file_content
        else file_content
    )

    if not isinstance(packages_list, IndexedList):
        return []

    for package in packages_list:
        parsed_package_with_dependencies = _get_parsed_package_and_dependencies(package, packages)
        if not parsed_package_with_dependencies:
            continue

        parsed_package, dependencies = parsed_package_with_dependencies
        if not parsed_package or not dependencies:
            continue

        raw_dependency_names = list(dependencies.keys()) if dependencies else []
        for dep_name in raw_dependency_names:
            dependency_parsed_package = next((x for x in packages if x.name == dep_name), None)
            if dependency_parsed_package:
                relationships.append(
                    Relationship(
                        from_=dependency_parsed_package.id_,
                        to_=parsed_package.id_,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships


def _is_dev_package(
    package: IndexedDict[str, ParsedValue], dev_packages: IndexedList[ParsedValue]
) -> bool:
    name_value = package.get("name")
    name = name_value if isinstance(name_value, str) else None

    return name in dev_packages if name else False


def _get_parsed_package_and_dependencies(
    package_data: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package_data, IndexedDict):
        return None

    package_name = package_data.get("name")
    if not isinstance(package_name, str):
        return None

    parsed_package = next((p for p in packages if p.name == package_name), None)

    raw_dependencies = package_data.get("require")
    if not parsed_package or not isinstance(raw_dependencies, IndexedDict):
        return None

    return parsed_package, raw_dependencies



import re

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.ruby.package_builder import new_gem_package
from labels.parsers.cataloger.utils import get_enriched_location

GEM_LOCK_DEP: re.Pattern[str] = re.compile(r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$")


def parse_gemfile_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = _collect_packages(reader.read_closer.read(), reader.location)

    return packages, []


def _collect_packages(file_content: str, location: Location) -> list[Package]:
    packages: list[Package] = []
    line_gem: bool = False

    for line_number, line in enumerate(file_content.splitlines(), 1):
        if line.startswith("GEM"):
            line_gem = True
        elif line_gem:
            if matched := GEM_LOCK_DEP.match(line):
                pkg_name = matched.group("gem")
                pkg_version = matched.group("version")

                new_location = get_enriched_location(
                    location,
                    line=line_number,
                    is_transitive=False,
                    package_manager=PackageManager.BUNDLER,
                )

                package = new_gem_package(pkg_name, pkg_version, new_location)
                if package:
                    packages.append(package)
            elif not line:
                break

    return packages



from collections.abc import Callable

from labels.model.package import Ecosystem, Package
from labels.model.syft_sbom import SyftArtifact
from labels.parsers.builder.alpine.builder_apk import builder as builder_apk_pkg
from labels.parsers.builder.arch.builder_alpm import builder as builder_alpm_pkg
from labels.parsers.builder.debian.builder_deb import builder as builder_deb_pkg
from labels.parsers.builder.generic_builder import builder as generic_builder
from labels.parsers.builder.java.builder_java import builder as builder_java_pkg
from labels.parsers.builder.python.builder_python import builder as builder_python_pkg
from labels.parsers.builder.redhat.builder_rpm import builder as builder_rpmdb_pkg

BUILDERS: dict[Ecosystem, Callable[[SyftArtifact, Ecosystem], Package | None]] = {
    Ecosystem.DEBIAN: builder_deb_pkg,
    Ecosystem.ALPINE: builder_apk_pkg,
    Ecosystem.ALPM: builder_alpm_pkg,
    Ecosystem.RPM: builder_rpmdb_pkg,
    Ecosystem.PYPI: builder_python_pkg,
    Ecosystem.MAVEN: builder_java_pkg,
    Ecosystem.GO: generic_builder,
    Ecosystem.NPM: generic_builder,
    Ecosystem.NUGET: generic_builder,
    Ecosystem.RUBYGEMS: generic_builder,
    Ecosystem.PACKAGIST: generic_builder,
    Ecosystem.PUB: generic_builder,
    Ecosystem.SWIFTURL: generic_builder,
}


def build_package(
    artifact: SyftArtifact,
    ecosystem: Ecosystem,
) -> Package | None:
    if builder := BUILDERS.get(ecosystem):
        return builder(artifact, ecosystem)
    return None



from collections.abc import Mapping
from pathlib import Path
from typing import cast

import yaml


def load_yaml_mapping(path: str) -> Mapping[str, object]:
    file_path = Path(path)
    with file_path.open(encoding="utf-8") as file:
        data = cast("object", yaml.safe_load(file))

    if data is None:
        return {}

    if not isinstance(data, Mapping):
        msg = "YAML root must be a mapping"
        raise TypeError(msg)

    return data



from pydantic import ValidationError

from labels.model.ecosystem_data.java import (
    JavaArchive,
    JavaManifest,
    JavaPomParent,
    JavaPomProject,
    JavaPomProperties,
)
from labels.model.package import Ecosystem, Package
from labels.model.syft_sbom import SyftArtifact, SyftManifestItem, SyftMetadata
from labels.parsers.builder.utils import get_language, get_locations
from labels.parsers.cataloger.utils import log_malformed_package_warning


def _parse_key_value_items(items: list[SyftManifestItem]) -> dict[str, str]:
    return {item.key: item.value for item in items if item.key and item.value}


def _get_java_manifest(metadata: SyftMetadata) -> JavaManifest | None:
    if not metadata.manifest:
        return None

    main = _parse_key_value_items(metadata.manifest.main or [])

    sections: list[dict[str, str]] = [
        section_dict
        for section in (metadata.manifest.sections or [])
        if (section_dict := _parse_key_value_items(section))
    ]

    return JavaManifest(
        main=main,
        sections=sections,
    )


def _get_java_pom_properties(metadata: SyftMetadata) -> JavaPomProperties | None:
    if not metadata.pom_properties:
        return None

    return JavaPomProperties(
        name=metadata.pom_properties.get("name"),
        group_id=metadata.pom_properties.get("groupId"),
        artifact_id=metadata.pom_properties.get("artifactId"),
        version=metadata.pom_properties.get("version"),
    )


def _get_java_pom_project(metadata: SyftMetadata) -> JavaPomProject | None:
    if not metadata.pom_project:
        return None

    parent: JavaPomParent | None = None

    if (
        (parent_data := metadata.pom_project.parent)
        and (group_id := parent_data.get("group_id"))
        and (artifact_id := parent_data.get("artifact_id"))
        and (version := parent_data.get("version"))
    ):
        parent = JavaPomParent(
            group_id=group_id,
            artifact_id=artifact_id,
            version=version,
        )

    return JavaPomProject(
        group_id=metadata.pom_project.group_id,
        artifact_id=metadata.pom_project.artifact_id,
        version=metadata.pom_project.version,
        name=metadata.pom_project.name,
        parent=parent,
    )


def _get_dpkg_entry(metadata: SyftMetadata) -> JavaArchive | None:
    manifest = _get_java_manifest(metadata)
    pom_properties = _get_java_pom_properties(metadata)
    pom_project = _get_java_pom_project(metadata)

    return JavaArchive(
        manifest=manifest,
        pom_properties=pom_properties,
        pom_project=pom_project,
    )


def _get_pkg_name(artifact_name: str, p_url: str) -> str | None:
    if not artifact_name or not p_url:
        return None

    p_url_splits = p_url.split("/")

    if len(p_url_splits) >= 3:
        group_id = p_url_splits[1]
        return f"{group_id}:{artifact_name}"

    return None


def builder(
    artifact: SyftArtifact,
    ecosystem: Ecosystem,
) -> Package | None:
    pkg_language = get_language(artifact.language)
    pkg_locations = get_locations(artifact.locations)

    metadata = artifact.metadata
    pkg_ecosystem_data = _get_dpkg_entry(metadata)

    p_url = artifact.p_url

    pkg_name = _get_pkg_name(artifact.name, p_url) or artifact.name

    try:
        package = Package(
            name=pkg_name,
            version=artifact.version,
            p_url=p_url,
            locations=pkg_locations,
            ecosystem=ecosystem,
            ecosystem_data=pkg_ecosystem_data,
            found_by=artifact.found_by,
            language=pkg_language,
            syft_id=artifact.id_,
        )
    except ValidationError as ex:
        log_malformed_package_warning(pkg_locations[0], ex)
        return None

    return package



from typing import NamedTuple

from labels.model.file import DependencyChain, DependencyType
from labels.model.package import Ecosystem, Package
from labels.model.relationship import Relationship


class PathContext(NamedTuple):
    rel_map: dict[str, list[str]]
    package_names_cache: dict[str, str]
    packages_by_id: dict[str, Package]
    current_name: str
    current_version: str


class TraversalState(NamedTuple):
    queue: list[tuple[str, list[str]]]
    visited: set[str]
    results: dict[str, list[str]]


class CurrentNode(NamedTuple):
    id: str
    path: list[str]


def calculate_top_parents_for_packages(
    packages: list[Package],
    relationships: list[Relationship],
) -> list[Package]:
    rel_map = build_relationship_map(relationships)

    package_names_cache = {pkg.id_: f"{pkg.name}@{pkg.version}" for pkg in packages}
    packages_by_id = {pkg.id_: pkg for pkg in packages}

    for pkg in packages:
        if pkg.ecosystem != Ecosystem.NPM:
            continue
        for location in pkg.locations:
            node_id = f"{pkg.id_}@{location.location_id()}"
            context = PathContext(
                rel_map=rel_map,
                package_names_cache=package_names_cache,
                packages_by_id=packages_by_id,
                current_name=pkg.name,
                current_version=pkg.version,
            )
            top_parents_with_paths = find_top_parents(node_id, context)
            current_pkg_name_version = f"{pkg.name}@{pkg.version}"

            if (
                location.dependency_type in (DependencyType.DIRECT, DependencyType.ROOT)
                and current_pkg_name_version in top_parents_with_paths
            ):
                del top_parents_with_paths[current_pkg_name_version]

            if top_parents_with_paths:
                location.top_parents = sorted(top_parents_with_paths.keys())
                location.dependency_chains = [
                    DependencyChain(depth=len(path), chain=path)
                    for path in top_parents_with_paths.values()
                ]
                location.dependency_chains.sort(key=lambda x: (x.depth, x.chain[-1]))
            else:
                location.top_parents = None
                location.dependency_chains = None

    return packages


def build_relationship_map(relationships: list[Relationship]) -> dict[str, list[str]]:
    rel_map: dict[str, list[str]] = {}
    for rel in relationships:
        if rel.type.value == "dependency-of":
            if rel.from_ not in rel_map:
                rel_map[rel.from_] = []
            rel_map[rel.from_].append(rel.to_)
    return rel_map


MAX_TRAVERSAL_ITERATIONS = 100_000


def _extract_package_id(node_id: str) -> str:
    return node_id.split("@")[0]


def _extract_location_id(node_id: str) -> str:
    return node_id.split("@", 1)[1]


def _get_package_name_version(package_id: str, cache: dict[str, str]) -> str:
    return cache.get(package_id, package_id)


def _get_dependency_type(
    node_id: str,
    packages_by_id: dict[str, Package],
) -> DependencyType | None:
    pkg_id = _extract_package_id(node_id)
    pkg = packages_by_id.get(pkg_id)

    if not pkg:
        return None

    location_id = _extract_location_id(node_id)
    location = next(
        (loc for loc in pkg.locations if loc.location_id() == location_id),
        None,
    )

    return location.dependency_type if location else None


def _should_add_as_top_parent(
    node_id: str,
    packages_by_id: dict[str, Package],
) -> bool:
    dep_type = _get_dependency_type(node_id, packages_by_id)

    if dep_type is None:
        return True

    return dep_type == DependencyType.DIRECT


def _is_root_node(node_id: str, packages_by_id: dict[str, Package]) -> bool:
    dep_type = _get_dependency_type(node_id, packages_by_id)
    return dep_type == DependencyType.ROOT


def _add_node_to_results(
    node_id: str,
    path: list[str],
    results: dict[str, list[str]],
    context: PathContext,
) -> None:
    pkg_id = _extract_package_id(node_id)
    pkg_name = _get_package_name_version(pkg_id, context.package_names_cache)

    if pkg_name not in results:
        results[pkg_name] = path


def _handle_circular_dependencies(
    parents: list[str],
    visited: set[str],
    current_path: list[str],
    results: dict[str, list[str]],
    context: PathContext,
) -> None:
    if not parents or not all(p in visited for p in parents):
        return

    if not context.packages_by_id:
        return

    for parent_id in parents:
        if not _should_add_as_top_parent(parent_id, context.packages_by_id):
            continue

        pkg_id = _extract_package_id(parent_id)
        parent_name = _get_package_name_version(pkg_id, context.package_names_cache)
        extended_path = [*current_path, parent_name]

        if parent_name not in results:
            results[parent_name] = extended_path


def _process_parents(
    parents: list[str],
    current: CurrentNode,
    state: TraversalState,
    context: PathContext,
) -> None:
    for parent_id in parents:
        if _is_root_node(parent_id, context.packages_by_id):
            _add_node_to_results(current.id, current.path, state.results, context)
            continue

        if parent_id in state.visited:
            continue

        parent_pkg_id = _extract_package_id(parent_id)
        parent_name = _get_package_name_version(parent_pkg_id, context.package_names_cache)

        state.visited.add(parent_id)
        state.queue.append((parent_id, [*current.path, parent_name]))


def find_top_parents(
    node_id: str,
    context: PathContext,
) -> dict[str, list[str]]:
    start_pkg_id = _extract_package_id(node_id)
    start_name = _get_package_name_version(start_pkg_id, context.package_names_cache)

    state = TraversalState(
        queue=[(node_id, [start_name])],
        visited={node_id},
        results={},
    )
    iterations = 0

    while state.queue:
        iterations += 1
        if iterations > MAX_TRAVERSAL_ITERATIONS:
            break

        current_id, current_path = state.queue.pop(0)
        current = CurrentNode(id=current_id, path=current_path)
        parents = context.rel_map.get(current_id, [])

        if not parents:
            if _should_add_as_top_parent(current_id, context.packages_by_id):
                _add_node_to_results(current_id, current_path, state.results, context)
            continue

        _process_parents(parents, current, state, context)
        _handle_circular_dependencies(parents, state.visited, current_path, state.results, context)

    return state.results



from cyclonedx.validation import ValidationError as CycloneDXError
from jsonschema import ValidationError as JSONSchemaError
from spdx_tools.spdx.validation.validation_message import ValidationMessage
from tree_sitter import Node

from labels.model.core import SourceType


class CustomBaseError(Exception):
    """Base exception class for custom exceptions."""


class AwsRoleNotDefinedError(CustomBaseError):
    """Exception raised when the AWS role is not defined."""

    def __init__(self) -> None:
        error_msg = "The AWS role wasn't defined in the configuration"
        super().__init__(error_msg)


class AwsExternalIdNotDefinedError(CustomBaseError):
    """Exception raised when the AWS External ID is not defined."""

    def __init__(self) -> None:
        error_msg = "The AWS External ID wasn't defined in the configuration"
        super().__init__(error_msg)


class AwsCredentialsNotFoundError(CustomBaseError):
    """Exception raised when AWS credentials are not found."""

    def __init__(self) -> None:
        error_msg = "The AWS credentials were not found or could not be retrieved."
        super().__init__(error_msg)


class CycloneDXValidationError(CustomBaseError):
    """Exception for CycloneDX validation errors."""

    def __init__(self, error: CycloneDXError) -> None:
        """Initialize the constructor."""
        header = "Exception - CycloneDx validation error\n"
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class DockerImageNotFoundError(CustomBaseError):
    """Exception raised when a Docker image is not found."""

    def __init__(self, image: str, extra_msg: str | None = None) -> None:
        error_msg = f"Docker image not found: {image}"
        if extra_msg:
            error_msg += f"\nError details: {extra_msg}"
        super().__init__(error_msg)


class DuplicatedKeyError(CustomBaseError):
    """Exception raised for duplicated keys."""

    def __init__(self, key: str) -> None:
        """Initialize the constructor."""
        super().__init__(f"Defining a key multiple times is invalid: {key}")


class FluidJSONValidationError(CustomBaseError):
    """Exception for Fluid JSON validation errors."""

    def __init__(self, error: JSONSchemaError) -> None:
        """Initialize the constructor."""
        header = "Exception - Fluid JSON validation error\n"
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class ForbiddenModuleImportedError(CustomBaseError):
    """Exception raised when a forbidden module is imported."""


class InvalidConfigFileError(CustomBaseError):
    """Exception raised for when the sbom config file is invalid."""

    def __init__(self, file_path: str, extra_msg: str | None = None) -> None:
        message = f"The configuration file is not a valid file: {file_path}"
        if extra_msg:
            message += f"\n{extra_msg}"
        super().__init__(message)


class InvalidDBFormatError(CustomBaseError):
    """Exception raised for invalid DB format."""


class InvalidImageReferenceError(CustomBaseError):
    """Exception raised when the image reference is invalid."""

    def __init__(self, image_ref: str) -> None:
        error_msg = f"The image reference '{image_ref}' is invalid."
        super().__init__(error_msg)


class InvalidMetadataError(CustomBaseError):
    """Exception raised for when metadata is invalid."""

    def __init__(self, error_message: str) -> None:
        message = error_message
        super().__init__(message)


class InvalidTypeError(CustomBaseError):
    """Exception raised for invalid types."""


class SyftNotFoundError(CustomBaseError):
    """Exception raised when the Syft binary is not found."""

    def __init__(self) -> None:
        error_msg = "The 'syft' binary was not found in the system."
        super().__init__(error_msg)


class SyftOutputParsingError(CustomBaseError):
    """Exception raised when Syft output cannot be parsed into a valid SBOM."""

    def __init__(self, image_ref: str) -> None:
        error_msg = f"Failed to parse Syft output for image: {image_ref}"
        super().__init__(error_msg)


class SPDXValidationError(CustomBaseError):
    """Exception for SPDX validation errors."""

    def __init__(self, error_messages: list[ValidationMessage]) -> None:
        """Initialize the constructor."""
        header = "Exception - SPDX validation error\n"
        error_details = "\n".join(
            (f"Validation error: {message.validation_message}\nContext: {message.context}")
            for message in error_messages
        )
        msg = f"❌ {header}{error_details}"
        super().__init__(msg)


class UnexpectedChildrenLengthError(CustomBaseError):
    """Exception raised for nodes with unexpected number of children."""

    def __init__(self, node: Node | str, expected_length: int) -> None:
        type_name = node.type if isinstance(node, Node) else node
        super().__init__(f"Unexpected node type {type_name} for {expected_length} children")


class UnexpectedExceptionError(CustomBaseError):
    """Exception for unexpected errors encountered during SBOM execution."""

    def __init__(self, error: Exception) -> None:
        """Initialize the constructor."""
        header = (
            "Exception - An unexpected exception was encountered "
            "during SBOM execution. The process will be terminated to prevent "
            "potential inconsistencies.\n"
        )
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class UnexpectedNodeError(CustomBaseError):
    """Exception raised for unexpected nodes."""

    def __init__(self, node: Node | str) -> None:
        type_name = node.type if isinstance(node, Node) else node
        if isinstance(node, Node) and node.text:
            value = node.text.decode("utf-8")
            super().__init__(
                f"Unexpected node type {type_name} with value {value}",
            )
        else:
            super().__init__(f"Unexpected node type {type_name}")


class UnexpectedNodeTypeError(CustomBaseError):
    """Exception raised for unexpected node types."""

    def __init__(self, node: Node | str, expected_type: str | None = None) -> None:
        type_name = node.type if isinstance(node, Node) else node
        if expected_type:
            super().__init__(f"Unexpected node type {type_name} for {expected_type}")
        else:
            super().__init__(f"Unexpected node type {type_name}")


class UnexpectedSBOMSourceError(CustomBaseError):
    """Exception raised when the SBOM source in the configuration is not recognized."""

    def __init__(self, source: SourceType) -> None:
        error_msg = f"Unrecognized SBOM source: {source}"
        super().__init__(error_msg)


class UnexpectedValueTypeError(CustomBaseError):
    """Exception raised for unexpected value types."""



import logging
from collections.abc import Callable

import reactivex
from reactivex import Observable
from reactivex.abc import ObserverBase
from reactivex.scheduler import ThreadPoolScheduler

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.parser import Task
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver

LOGGER = logging.getLogger(__name__)


def execute_parsers(
    resolver: Resolver,
    environment: Environment,
) -> Callable[[Observable[Task]], Observable[tuple[list[Package], list[Relationship]]]]:
    def _handle(source: Observable[Task]) -> Observable[tuple[list[Package], list[Relationship]]]:
        def subscribe(
            observer: ObserverBase[tuple[list[Package], list[Relationship]]],
            scheduler: ThreadPoolScheduler | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Task) -> None:
                LOGGER.debug("Working on %s", value.location.access_path)
                content_reader = resolver.file_contents_by_location(value.location)

                try:
                    if content_reader is not None and (
                        result := value.parser(
                            resolver,
                            environment,
                            LocationReadCloser(location=value.location, read_closer=content_reader),
                        )
                    ):
                        discover_packages, relationships = result
                        for pkg in discover_packages:
                            pkg.found_by = value.parser_name
                        observer.on_next((discover_packages, relationships))
                except Exception as ex:  # noqa: BLE001
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)  # type: ignore[arg-type]

    return _handle



from labels.model.ecosystem_data.alpine import ApkDBEntry
from labels.model.ecosystem_data.debian import DpkgDBEntry
from labels.model.package import Ecosystem, Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.syft_sbom import SyftSBOM

VERSION_SPECIFIERS = "[(<>="
CHOICE_SEPARATOR = "|"


def _strip_version_specifier(item: str) -> str:
    index = next((i for i, char in enumerate(item) if char in VERSION_SPECIFIERS), None)

    if not index:
        return item.strip()

    return item[:index].strip()


def _split_package_choices(value: str) -> list[str]:
    fields = value.split(CHOICE_SEPARATOR)
    return [_strip_version_specifier(field) for field in fields if field.strip()]


def _build_package_lookup(packages: list[Package]) -> dict[str, list[Package]]:
    lookup: dict[str, list[Package]] = {}

    for package in packages:
        if not isinstance(package.ecosystem_data, (DpkgDBEntry, ApkDBEntry)):
            continue

        if package.name:
            lookup.setdefault(package.name, []).append(package)

        ecosystem_data = package.ecosystem_data
        for provides in ecosystem_data.provides or []:
            key = _strip_version_specifier(provides)
            lookup.setdefault(key, []).append(package)

    return lookup


def _get_os_package_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    lookup_table = _build_package_lookup(packages)

    for package_to in packages:
        if not isinstance(package_to.ecosystem_data, (DpkgDBEntry, ApkDBEntry)):
            continue

        ecosystem_data = package_to.ecosystem_data

        all_dependencies: list[str] = list(ecosystem_data.dependencies or [])

        if isinstance(ecosystem_data, DpkgDBEntry) and ecosystem_data.pre_dependencies:
            all_dependencies.extend(ecosystem_data.pre_dependencies)

        for dep_specifier in all_dependencies:
            dependency_names = _split_package_choices(dep_specifier)

            for dependency in dependency_names:
                relationships.extend(
                    [
                        Relationship(
                            from_=package_from.id_,
                            to_=package_to.id_,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        )
                        for package_from in lookup_table.get(dependency, [])
                        if package_from.id_ != package_to.id_
                    ]
                )
    return relationships


def _get_pkg_manager_relationships(
    docker_sbom: SyftSBOM, packages: list[Package]
) -> list[Relationship]:
    relationships: list[Relationship] = []
    pkg_map = {pkg.syft_id: pkg for pkg in packages}

    for relationship in docker_sbom.relationships:
        if relationship.type != RelationshipType.DEPENDENCY_OF_RELATIONSHIP.value:
            continue

        from_pkg = pkg_map.get(relationship.parent)
        to_pkg = pkg_map.get(relationship.child)

        if from_pkg and to_pkg and from_pkg.id_ != to_pkg.id_:
            relationships.append(
                Relationship(
                    from_=from_pkg.id_,
                    to_=to_pkg.id_,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
            )

    return relationships


def get_relationships(docker_sbom: SyftSBOM, packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    grouped_packages: dict[str, list[Package]] = {}

    for package in packages:
        if package.ecosystem in [Ecosystem.ALPINE, Ecosystem.DEBIAN]:
            grouped_packages.setdefault("os", []).append(package)

        if package.ecosystem in [
            Ecosystem.NUGET,
            Ecosystem.PACKAGIST,
            Ecosystem.SWIFTURL,
        ]:
            grouped_packages.setdefault("pkg_manager", []).append(package)

    if os_packages := grouped_packages.get("os"):
        relationships.extend(_get_os_package_relationships(os_packages))

    if pkg_manager_packages := grouped_packages.get("pkg_manager"):
        relationships.extend(_get_pkg_manager_relationships(docker_sbom, pkg_manager_packages))

    return relationships



import logging

from pydantic import ValidationError

from labels.advisories import images as images_advisories
from labels.advisories import roots as roots_advisories
from labels.model.advisories import Advisory
from labels.model.ecosystem_data.alpine import ApkDBEntry
from labels.model.ecosystem_data.arch import AlpmDBEntry
from labels.model.ecosystem_data.debian import DpkgDBEntry
from labels.model.ecosystem_data.redhat import RpmDBEntry
from labels.model.package import Ecosystem, Package
from labels.parsers.cataloger.utils import extract_distro_info
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


ROOT_TYPES: set[Ecosystem] = {
    Ecosystem.NPM,
    Ecosystem.PUB,
    Ecosystem.NUGET,
    Ecosystem.MAVEN,
    Ecosystem.PACKAGIST,
    Ecosystem.PYPI,
    Ecosystem.RUBYGEMS,
    Ecosystem.GO,
    Ecosystem.SWIFTURL,
}

IMAGES_TYPES: set[Ecosystem] = {
    Ecosystem.ALPM,
    Ecosystem.DEBIAN,
    Ecosystem.ALPINE,
    Ecosystem.RPM,
}


def update_root_advisories(package: Package) -> list[Advisory]:
    safe_versions = roots_advisories.get_safe_versions(package.ecosystem, package.name)
    package.safe_versions = safe_versions
    return roots_advisories.get_vulnerabilities(
        package.ecosystem,
        package.name,
        package.version,
        safe_versions,
    )


def _get_upstream_info(package: Package) -> tuple[str | None, str | None]:
    if (
        package.ecosystem == Ecosystem.ALPINE
        and isinstance(package.ecosystem_data, ApkDBEntry)
        and package.ecosystem_data
        and package.ecosystem_data.origin_package
    ):
        return package.ecosystem_data.origin_package, package.version

    if (
        package.ecosystem == Ecosystem.DEBIAN
        and isinstance(package.ecosystem_data, DpkgDBEntry)
        and package.ecosystem_data
        and package.ecosystem_data.source
    ):
        return (
            package.ecosystem_data.source,
            package.ecosystem_data.source_version or package.version,
        )

    if (
        package.ecosystem == Ecosystem.ALPM
        and isinstance(package.ecosystem_data, AlpmDBEntry)
        and package.ecosystem_data
        and package.ecosystem_data.base_package
    ):
        return package.ecosystem_data.base_package, package.version

    if (
        package.ecosystem == Ecosystem.RPM
        and isinstance(package.ecosystem_data, RpmDBEntry)
        and package.ecosystem_data
        and package.ecosystem_data.source_rpm
    ):
        return _extract_rpm_upstream_info(package.ecosystem_data.source_rpm)

    return None, None


def _extract_rpm_upstream_info(source_rpm: str) -> tuple[str | None, str | None]:
    """Extract upstream package name and version from RPM source package.

    RPM format: NAME-VERSION-RELEASE[.ARCH].src.rpm or NAME-VERSION-RELEASE[.ARCH].nosrc.rpm
    Example: pam-1.5.1-25.el9_6.x86_64.src.rpm -> (pam, 1.5.1-25.el9_6)
    Example: pam-1.5.1-25.el9_6.src.rpm -> (pam, 1.5.1-25.el9_6)
    Example: foo-1-1.el7.nosrc.rpm -> (foo, 1-1.el7.nosrc)
    """
    # Remove source RPM suffixes
    if source_rpm.endswith(".nosrc.rpm"):
        source_rpm = source_rpm[:-10]  # Remove .nosrc.rpm
    elif source_rpm.endswith(".src.rpm"):
        source_rpm = source_rpm[:-8]  # Remove .src.rpm

    # Known RPM architectures - only remove if last component matches these
    known_architectures = {
        "x86_64",
        "i386",
        "i486",
        "i586",
        "i686",
        "aarch64",
        "armv7hl",
        "armv7l",
        "armv6hl",
        "ppc64le",
        "ppc64",
        "s390x",
        "noarch",
    }

    # Remove architecture only if present and recognized
    if "." in source_rpm:
        possible_arch = source_rpm.rsplit(".", 1)[1]
        if possible_arch in known_architectures:
            source_rpm = source_rpm.rsplit(".", 1)[0]

    # Now split by hyphens: NAME-VERSION-RELEASE
    parts = source_rpm.split("-")
    if len(parts) >= 3:
        name = "-".join(parts[:-2])
        version = parts[-2]
        release = parts[-1]
        full_version = f"{version}-{release}"
        return name, full_version

    return None, None


def _get_rhel_distro_info(package: Package) -> tuple[str | None, str | None]:
    """Extract RHEL distribution information from RPM package.

    Returns:
        tuple: (distro_id, distro_version) where:
        - distro_id: "rpm" for RHEL packages
        - distro_version: "rhel9.6" for RHEL 9.6, "rhel8" for RHEL 8, etc.

    """
    if not (
        package.ecosystem == Ecosystem.RPM
        and isinstance(package.ecosystem_data, RpmDBEntry)
        and package.ecosystem_data
        and package.ecosystem_data.release
    ):
        return None, None

    release = package.ecosystem_data.release

    # Look for RHEL pattern: .elX or .elX_Y
    if ".el" not in release:
        return None, None

    # Extract the part after .el (e.g., "9_6" from "63.el9_6")
    el_part = release.split(".el")[-1]
    if not el_part or not el_part[0].isdigit():
        return None, None
    # Parse RHEL version
    if "_" in el_part:
        # Full version: "9_6.1" -> "rhel9.6.1"
        parts = el_part.split("_")
        major, minor = parts[0], parts[1]
        return "rpm", f"rhel{major}.{minor}"

    # Major version only: "9" -> "rhel9"
    return "rpm", f"rhel{el_part}"


def _normalize_rpm_version(version: str) -> str:
    """Normalize RPM version by adding epoch 0: if not present."""
    # Check if version already has an epoch (contains ':')
    if ":" in version:
        return version

    # Add epoch 0: if not present
    return f"0:{version}"


def _get_upstream_info_if_different(
    package: Package, upstream_package: str | None, upstream_version: str | None
) -> tuple[str | None, str | None] | None:
    if (
        upstream_package
        and upstream_version
        and (upstream_package != package.name or upstream_version != package.version)
    ):
        return (upstream_package, upstream_version)
    return None


def update_image_advisories(package: Package) -> list[Advisory]:
    distro_id = None
    distro_version = None
    if package.ecosystem in [Ecosystem.DEBIAN, Ecosystem.ALPM, Ecosystem.ALPINE]:
        distro_id, distro_version, _ = extract_distro_info(package.p_url)
        distro_version = (
            "v" + ".".join(str(distro_version).split(".")[0:2])
            if package.ecosystem == Ecosystem.ALPINE
            else str(distro_version)
        )

    if package.ecosystem == Ecosystem.RPM:
        distro_id, distro_version = _get_rhel_distro_info(package)

    upstream_package, upstream_version = _get_upstream_info(package)
    upstream_info = _get_upstream_info_if_different(package, upstream_package, upstream_version)

    normalized_version = package.version
    if package.ecosystem == Ecosystem.RPM:
        normalized_version = _normalize_rpm_version(package.version)

    return images_advisories.get_vulnerabilities(
        str(distro_id),
        package.name,
        normalized_version,
        distro_version,
        upstream_info,
    )


def get_package_advisories(package: Package) -> list[Advisory] | None:
    try:
        pkg_advisories = []
        if package.ecosystem in ROOT_TYPES:
            pkg_advisories = update_root_advisories(package)
        if package.ecosystem in IMAGES_TYPES:
            pkg_advisories = update_image_advisories(package)
    except ValidationError as ex:
        LOGGER.exception(
            "Unable to complete package advisories",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": package.locations,
                    "ecosystem": package.ecosystem,
                },
            },
        )
        return None
    return pkg_advisories


def complete_package_advisories(package: Package) -> Package:
    if pkg_advisories := get_package_advisories(package):
        package.advisories = pkg_advisories
    return package



from typing import cast

from bs4 import BeautifulSoup, Tag

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.utils import get_enriched_location

PACKAGE_TAG = "package"


def parse_dotnet_pkgs_config(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = BeautifulSoup(reader.read_closer.read(), features="html.parser")

    packages = _collect_packages(file_content, reader.location)

    return packages, []


def _collect_packages(file_content: BeautifulSoup, location: Location) -> list[Package]:
    packages: list[Package] = []

    packages_tags: list[Tag] = file_content.find_all(PACKAGE_TAG, recursive=True)
    for raw_package in packages_tags:
        name = cast("str | None", raw_package.get("id"))
        version = cast("str | None", raw_package.get("version"))

        new_location = get_enriched_location(
            location,
            line=raw_package.sourceline,
            is_transitive=False,
            package_manager=PackageManager.NUGET,
        )

        new_package = new_dotnet_package(name, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages



import logging

import tree_sitter_json
from tree_sitter import Language as TLanguage
from tree_sitter import Node, Parser

from labels.model.indexables import IndexedDict, IndexedList, ParsedValue

LOGGER = logging.getLogger(__name__)


def _handle_array_node(node: Node) -> tuple[Node, IndexedList[ParsedValue]]:
    data: IndexedList[ParsedValue] = IndexedList(node)
    for child in node.children:
        if child.type not in ("[", "]", ","):
            if not (parsed_node := handle_json_node(child)):
                continue
            value_node, value = parsed_node
            data.append((value, value_node))
    return node, data


def _handle_object_node(node: Node) -> tuple[Node, IndexedDict[str, ParsedValue]]:
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for child in node.children:
        if child.type == "pair":
            key_n, _, value_n = child.children
            if not key_n.text:  # pragma: no cover
                continue
            key = key_n.text[1:-1].decode("utf-8")
            if not (parsed_node := handle_json_node(value_n)):  # pragma: no cover
                continue
            value_node, value_value = parsed_node
            data[(key, key_n)] = (value_value, value_node)
    return node, data


def _parse_json_number(node: Node) -> int | float:
    node_value = node.text.decode("utf-8") if node.text else "0"
    try:
        return int(node_value)
    except ValueError:
        return float(node_value)


def handle_json_node(node: Node) -> tuple[Node, ParsedValue] | None:
    value: tuple[Node, ParsedValue]
    match node.type:
        case "array":
            value = _handle_array_node(node)
        case "object":
            value = _handle_object_node(node)
        case "string":
            node_value = node.text[1:-1].decode("utf-8") if node.text else ""
            value = node, node_value
        case "number":
            value = node, _parse_json_number(node)
        case "true":
            value = node, True
        case "false":
            value = node, False
        case "null":
            value = node, None
        case _:
            LOGGER.debug("Ignoring unexpected node type encountered: %s", node.type)
            return None
    return value


def parse_json_with_tree_sitter(
    json: str,
) -> ParsedValue:
    parser_language = TLanguage(tree_sitter_json.language())
    parser = Parser(parser_language)
    result = parser.parse(json.encode("utf-8"))
    value: ParsedValue = IndexedDict()
    for child in result.root_node.children:
        if not (parsed_node := handle_json_node(child)):
            continue
        value = parsed_node[1]

    if json and (value is None or not isinstance(value, IndexedDict | IndexedList)):
        LOGGER.error("JSON parsing failed.", extra={"extra": {"json": json}})
        return IndexedDict()
    return value



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.metadata import HealthMetadata
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning
from labels.utils.strings import normalize_name


def new_dotnet_package(
    name: str | None,
    version: str | None,
    new_location: Location,
    health_metadata: HealthMetadata | None = None,
) -> Package | None:
    if not name or not version:
        return None

    normalized_package_name = normalize_name(name, Ecosystem.NUGET)
    p_url = PackageURL(type="nuget", name=normalized_package_name, version=version).to_string()

    try:
        return Package(
            name=normalized_package_name,
            version=version,
            locations=[new_location],
            language=Language.DOTNET,
            ecosystem=Ecosystem.NUGET,
            p_url=p_url,
            health_metadata=health_metadata,
        )
    except ValidationError as ex:
        log_malformed_package_warning(new_location, ex)

    return None



from os_release.parser import OsReleaseParseException, parse_str

from labels.model.ecosystem_data.python import WheelEggEcosystemData
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Release
from labels.model.resolver import Resolver


def identify_release(resolver: Resolver) -> Release | None:
    possible_files = [
        "/etc/os-release",
        "/usr/lib/os-release",
        "/etc/system-release-cpe",
        "/etc/redhat-release",
        "/bin/busybox",
    ]

    for file in possible_files:
        if not resolver.has_path(file):
            continue
        location = resolver.files_by_path(file)[0]
        content_reader = resolver.file_contents_by_location(location)
        if not content_reader:
            continue
        content = content_reader.read()
        release = parse_os_release(content)
        if release:
            return release
    return None


def parse_os_release(content: str) -> Release | None:
    try:
        release = parse_str(content)
    except OsReleaseParseException:
        release = _force_parse(content)
    if release:
        id_like: list[str] = []
        if "ID_LIKE" in release:
            id_like = sorted(release["ID_LIKE"].split(" "))
        return Release(
            pretty_name=release.get("PRETTY_NAME", ""),
            name=release.get("NAME", ""),
            id_=release.get("ID", ""),
            id_like=id_like,
            version=release.get("VERSION", ""),
            version_id=release.get("VERSION_ID", ""),
            version_code_name=release.get("VERSION_CODENAME", ""),
            build_id=release.get("BUILD_ID", ""),
            image_id=release.get("IMAGE_ID", ""),
            image_version=release.get("IMAGE_VERSION", ""),
            variant=release.get("VARIANT", ""),
            variant_id=release.get("VARIANT_ID", ""),
            home_url=release.get("HOME_URL", ""),
            support_url=release.get("SUPPORT_URL", ""),
            bug_report_url=release.get("BUG_REPORT_URL", ""),
            privacy_policy_url=release.get("PRIVACY_POLICY_URL", ""),
            cpe_name=release.get("CPE_NAME", ""),
            support_end=release.get("SUPPORT_END", ""),
        )
    return None


def _is_valid_key(key: str) -> bool:
    return key.isidentifier() and key.isupper()


def _force_parse(content: str) -> dict[str, str] | None:
    lines: list[tuple[str, str]] = []

    for raw_line in content.splitlines():
        line = raw_line.strip()
        if not line or "=" not in line:
            continue

        key, value = line.split("=", 1)
        key = key.strip()
        value = value.strip("\"'")

        if not _is_valid_key(key):
            continue

        lines.append((key, value))

    if not lines:
        return None

    return dict(lines)


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next((i for i, char in enumerate(item) if char in specifiers), None)

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def handle_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package in packages:
        if package.found_by == "python-installed-package-cataloger":
            if not isinstance(package.ecosystem_data, WheelEggEcosystemData):
                continue

            ecosystem_data = package.ecosystem_data
            for dep in ecosystem_data.dependencies if ecosystem_data.dependencies else []:
                dep_name = strip_version_specifier(dep)
                if dep_package := next((x for x in packages if x.name == dep_name), None):
                    relationships.append(
                        Relationship(
                            from_=dep_package.id_,
                            to_=package.id_,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )
    return relationships



from typing import Annotated

from pydantic import StringConstraints

NonEmptyStr = Annotated[str, StringConstraints(min_length=1)]
TrimmedNonEmptyStr = Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)]



from typing import Any, TypedDict


class FluidSBOM(TypedDict):
    sbom_details: dict[str, str | None]
    packages: list[dict[str, Any]]  # type: ignore[explicit-any]
    relationships: list[dict[str, str | list[str]]]



from collections import defaultdict

from cyclonedx.model.bom import Bom
from cyclonedx.model.bom_ref import BomRef
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.model.tool import Tool
from packageurl import PackageURL

from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.output.cyclonedx.complete_file import (
    add_component_properties,
    add_integrity,
    add_vulnerabilities,
)


def pkg_to_component(package: Package) -> Component:
    health_metadata = package.health_metadata
    integrity = add_integrity(health_metadata) if health_metadata else []
    properties = add_component_properties(package)

    return Component(
        type=ComponentType.LIBRARY,
        name=package.name,
        version=package.version,
        bom_ref=f"{package.name}@{package.version}",
        purl=PackageURL.from_string(package.p_url),
        properties=properties,
        hashes=integrity,
    )


def create_bom(namespace: str, version: str | None) -> Bom:
    bom = Bom()
    bom.metadata.component = Component(
        name=namespace,
        type=ComponentType.APPLICATION,
        bom_ref=BomRef(f"{namespace}@{version}" if version else namespace),
        version=version,
    )
    bom.metadata.tools.tools.add(Tool(vendor="Fluid Attacks", name="Fluid-Labels"))
    return bom


def add_components_to_bom(bom: Bom, component_cache: dict[str, Component]) -> None:
    for component in component_cache.values():
        bom.components.add(component)


def add_advisories_to_bom(
    bom: Bom,
    packages: list[Package],
) -> None:
    for package in packages:
        if package.advisories:
            vulnerabilities = add_vulnerabilities(package)
            for vulnerability in vulnerabilities:
                bom.vulnerabilities.add(vulnerability)


def add_relationships_to_bom(
    bom: Bom,
    relationships: list[Relationship],
    component_cache: dict[str, Component],
) -> None:
    dependency_map: dict[Component, list[Component]] = defaultdict(list)
    for relationship in relationships:
        if relationship.type.value == "dependency-of":
            # Extract base IDs (remove @location_id suffix if present)
            to_id = relationship.to_.split("@")[0] if "@" in relationship.to_ else relationship.to_
            from_id = (
                relationship.from_.split("@")[0]
                if "@" in relationship.from_
                else relationship.from_
            )

            to_pkg = component_cache.get(to_id)
            from_pkg = component_cache.get(from_id)
            if to_pkg and from_pkg:
                dependency_map[to_pkg].append(from_pkg)

    for ref, depends_on_list in dependency_map.items():
        bom.register_dependency(ref, depends_on_list)

    if bom.metadata.component:
        bom.register_dependency(bom.metadata.component, component_cache.values())
